#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilties to assist testing the Multiscene functionality.

Creating fake test data for use in the other Multiscene test modules.
"""


import dask.array as da
import numpy as np
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.dataset.dataid import ModifierTuple, WavelengthRange

DEFAULT_SHAPE = (5, 10)

local_id_keys_config = {"name": {
    "required": True,
},
    "wavelength": {
    "type": WavelengthRange,
},
    "resolution": None,
    "calibration": {
    "enum": [
        "reflectance",
        "brightness_temperature",
        "radiance",
        "counts"
    ]
},
    "polarization": None,
    "level": None,
    "modifiers": {
    "required": True,
    "default": ModifierTuple(),
    "type": ModifierTuple,
},
}


def _fake_get_enhanced_image(img, enhance=None, overlay=None, decorate=None):
    from trollimage.xrimage import XRImage
    return XRImage(img)


def _create_test_area(proj_str=None, shape=DEFAULT_SHAPE, extents=None):
    """Create a test area definition."""
    if proj_str is None:
        proj_str = "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. " \
                   "+lat_0=25 +lat_1=25 +units=m +no_defs"
    extents = extents or (-1000., -1500., 1000., 1500.)

    return AreaDefinition(
        "test",
        "test",
        "test",
        proj_str,
        shape[1],
        shape[0],
        extents
    )


def _create_test_int8_dataset(name, shape=DEFAULT_SHAPE, area=None, values=None, dims=("y", "x")):
    """Create a test DataArray object."""
    return xr.DataArray(
        da.ones(shape, dtype=np.uint8, chunks=shape) * values, dims=dims,
        attrs={"_FillValue": 255,
               "valid_range": [1, 15],
               "name": name, "area": area, "_satpy_id_keys": local_id_keys_config})


def _create_test_dataset(name, shape=DEFAULT_SHAPE, area=None, values=None, dims=("y", "x")):
    """Create a test DataArray object."""
    if values:
        return xr.DataArray(
            da.ones(shape, dtype=np.float32, chunks=shape) * values, dims=dims,
            attrs={"name": name, "area": area, "_satpy_id_keys": local_id_keys_config})

    return xr.DataArray(
        da.zeros(shape, dtype=np.float32, chunks=shape), dims=dims,
        attrs={"name": name, "area": area, "_satpy_id_keys": local_id_keys_config})


def _create_test_scenes(num_scenes=2, shape=DEFAULT_SHAPE, area=None):
    """Create some test scenes for various test cases."""
    from satpy import Scene
    ds1 = _create_test_dataset("ds1", shape=shape, area=area)
    ds2 = _create_test_dataset("ds2", shape=shape, area=area)
    scenes = []
    for _ in range(num_scenes):
        scn = Scene()
        scn["ds1"] = ds1.copy()
        scn["ds2"] = ds2.copy()
        scenes.append(scn)
    return scenes



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""SEVIRI specific compositors."""

from __future__ import annotations

import logging

from .core import GenericCompositor, IncompatibleAreas

LOG = logging.getLogger(__name__)


class RealisticColors(GenericCompositor):
    """Create a realistic colours composite for SEVIRI."""

    def __call__(self, projectables, *args, **kwargs):
        """Generate the composite."""
        projectables = self.match_data_arrays(projectables)
        vis06 = projectables[0]
        vis08 = projectables[1]
        hrv = projectables[2]

        try:
            ch3 = 3.0 * hrv - vis06 - vis08
            ch3.attrs = hrv.attrs
        except ValueError:
            raise IncompatibleAreas("Areas do not match")

        ndvi = (vis08 - vis06) / (vis08 + vis06)
        ndvi = ndvi.where(ndvi >= 0.0, 0.0)

        ch1 = ndvi * vis06 + (1.0 - ndvi) * vis08
        ch1.attrs = vis06.attrs
        ch2 = ndvi * vis08 + (1.0 - ndvi) * vis06
        ch2.attrs = vis08.attrs

        res = super(RealisticColors, self).__call__((ch1, ch2, ch3),
                                                    *args, **kwargs)
        return res



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit testing the enhancements functions, e.g. cira_stretch."""

import os
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr


@pytest.fixture
def fake_area():
    """Return a fake 2×2 area."""
    from pyresample.geometry import create_area_def
    return create_area_def("wingertsberg", 4087, area_extent=[-2_000, -2_000, 2_000, 2_000], shape=(2, 2))


_nwcsaf_geo_props = {
    "cma_geo": ("geo", "cma", None, "cma_pal", None, "cloudmask", "CMA", "uint8"),
    "cma_pps": ("pps", "cma", None, "cma_pal", None, "cloudmask", "CMA", "uint8"),
    "cma_extended_pps": ("pps", "cma_extended", None, "cma_extended_pal", None,
                         "cloudmask_extended", "CMA", "uint8"),
    "cmaprob_pps": ("pps", "cmaprob", None, "cmaprob_pal", None, "cloudmask_probability",
                    "CMAPROB", "uint8"),
    "ct_geo": ("geo", "ct", None, "ct_pal", None, "cloudtype", "CT", "uint8"),
    "ct_pps": ("pps", "ct", None, "ct_pal", None, "cloudtype", "CT", "uint8"),
    "ctth_alti_geo": ("geo", "ctth_alti", None, "ctth_alti_pal", None, "cloud_top_height",
                      "CTTH", "float64"),
    "ctth_alti_pps": ("pps", "ctth_alti", None, "ctth_alti_pal", "ctth_status_flag",
                      "cloud_top_height", "CTTH", "float64"),
    "ctth_pres_geo": ("geo", "ctth_pres", None, "ctth_pres_pal", None, "cloud_top_pressure",
                      "CTTH", "float64"),
    "ctth_pres_pps": ("pps", "ctth_pres", None, "ctth_pres_pal", None, "cloud_top_pressure",
                      "CTTH", "float64"),
    "ctth_tempe_geo": ("geo", "ctth_tempe", None, "ctth_tempe_pal", None, "cloud_top_temperature",
                       "CTTH", "float64"),
    "ctth_tempe_pps": ("pps", "ctth_tempe", None, "ctth_tempe_pal", None, "cloud_top_temperature",
                       "CTTH", "float64"),
    "cmic_phase_geo": ("geo", "cmic_phase", None, "cmic_phase_pal", None, "cloud_top_phase",
                       "CMIC", "uint8"),
    "cmic_phase_pps": ("pps", "cmic_phase", None, "cmic_phase_pal", "cmic_status_flag", "cloud_top_phase",
                       "CMIC", "uint8"),
    "cmic_reff_geo": ("geo", "cmic_reff", None, "cmic_reff_pal", None, "cloud_drop_effective_radius",
                      "CMIC", "float64"),
    "cmic_reff_pps": ("pps", "cmic_reff", "cmic_cre", "cmic_cre_pal", "cmic_status_flag",
                      "cloud_drop_effective_radius", "CMIC", "float64"),
    "cmic_cot_geo": ("geo", "cmic_cot", None, "cmic_cot_pal", None, "cloud_optical_thickness",
                     "CMIC", "float64"),
    "cmic_cot_pps": ("pps", "cmic_cot", None, "cmic_cot_pal", None, "cloud_optical_thickness",
                     "CMIC", "float64"),
    "cmic_cwp_pps": ("pps", "cmic_cwp", None, "cmic_cwp_pal", None, "cloud_water_path",
                     "CMIC", "float64"),
    "cmic_lwp_geo": ("geo", "cmic_lwp", None, "cmic_lwp_pal", None, "cloud_liquid_water_path",
                     "CMIC", "float64"),
    "cmic_lwp_pps": ("pps", "cmic_lwp", None, "cmic_lwp_pal", None, "liquid_water_path",
                     "CMIC", "float64"),
    "cmic_iwp_geo": ("geo", "cmic_iwp", None, "cmic_iwp_pal", None, "cloud_ice_water_path",
                     "CMIC", "float64"),
    "cmic_iwp_pps": ("pps", "cmic_iwp", None, "cmic_iwp_pal", None, "ice_water_path",
                     "CMIC", "float64"),
    "pc": ("geo", "pc", None, "pc_pal", None, "precipitation_probability", "PC", "uint8"),
    "crr": ("geo", "crr", None, "crr_pal", None, "convective_rain_rate", "CRR", "uint8"),
    "crr_accum": ("geo", "crr_accum", None, "crr_pal", None,
                  "convective_precipitation_hourly_accumulation", "CRR", "uint8"),
    "ishai_tpw": ("geo", "ishai_tpw", None, "ishai_tpw_pal", None, "total_precipitable_water",
                  "iSHAI", "float64"),
    "ishai_shw": ("geo", "ishai_shw", None, "ishai_shw_pal", None, "showalter_index",
                  "iSHAI", "float64"),
    "ishai_li": ("geo", "ishai_li", None, "ishai_li_pal", None, "lifted_index",
                 "iSHAI", "float64"),
    "ci_prob30": ("geo", "ci_prob30", None, "ci_pal", None, "convection_initiation_prob30",
                  "CI", "float64"),
    "ci_prob60": ("geo", "ci_prob60", None, "ci_pal", None, "convection_initiation_prob60",
                  "CI", "float64"),
    "ci_prob90": ("geo", "ci_prob90", None, "ci_pal", None, "convection_initiation_prob90",
                  "CI", "float64"),
    "asii_turb_trop_prob": ("geo", "asii_turb_trop_prob", None, "asii_turb_prob_pal", None,
                            "asii_prob", "ASII-NG", "float64"),
    "MapCellCatType": ("geo", "MapCellCatType", None, "MapCellCatType_pal", None,
                       "rdt_cell_type", "RDT-CW", "uint8"),
}


@pytest.mark.parametrize(
    "data",
    ["cma_geo", "cma_pps", "cma_extended_pps", "cmaprob_pps", "ct_geo",
     "ct_pps", "ctth_alti_geo", "ctth_alti_pps", "ctth_pres_geo",
     "ctth_pres_pps", "ctth_tempe_geo", "ctth_tempe_pps",
     "cmic_phase_geo", "cmic_phase_pps", "cmic_reff_geo",
     "cmic_reff_pps", "cmic_cot_geo", "cmic_cot_pps", "cmic_cwp_pps",
     "cmic_lwp_geo", "cmic_lwp_pps", "cmic_iwp_geo", "cmic_iwp_pps",
     "pc", "crr", "crr_accum", "ishai_tpw", "ishai_shw", "ishai_li",
     "ci_prob30", "ci_prob60", "ci_prob90", "asii_turb_trop_prob",
     "MapCellCatType"]
)
def test_nwcsaf_comps(fake_area, tmp_path, data):
    """Test loading NWCSAF composites."""
    from satpy import Scene
    from satpy.enhancements.enhancer import get_enhanced_image
    (flavour, dvname, altname, palettename, statusname, comp, filelabel, dtp) = _nwcsaf_geo_props[data]
    rng = (0, 100) if dtp == "uint8" else (-100, 1000)
    if flavour == "geo":
        fn = f"S_NWC_{filelabel:s}_MSG2_MSG-N-VISIR_20220124T094500Z.nc"
        reader = "nwcsaf-geo"
        id_ = {"satellite_identifier": "MSG4"}
    else:
        fn = f"S_NWC_{filelabel:s}_noaa20_00000_20230301T1200213Z_20230301T1201458Z.nc"
        reader = "nwcsaf-pps_nc"
        id_ = {"platform": "NOAA-20"}
    fk = tmp_path / fn
    # create a minimally fake netCDF file, otherwise satpy won't load the
    # composite
    ds = xr.Dataset(
        coords={"nx": [0], "ny": [0]},
        attrs={
            "source": "satpy unit test",
            "time_coverage_start": "0001-01-01T00:00:00Z",
            "time_coverage_end": "0001-01-01T01:00:00Z",
        }
    )
    ds.attrs.update(id_)
    ds.to_netcdf(fk)
    sc = Scene(filenames=[os.fspath(fk)], reader=[reader])
    sc[palettename] = xr.DataArray(
        da.tile(da.arange(256), [3, 1]).T,
        dims=("pal02_colors", "pal_RGB"))
    fake_alti = da.linspace(rng[0], rng[1], 4, chunks=2, dtype=dtp).reshape(2, 2)
    ancvars = [sc[palettename]]
    if statusname is not None:
        sc[statusname] = xr.DataArray(
            da.zeros(shape=(2, 2), dtype="uint8"),
            attrs={
                "area": fake_area,
                "_FillValue": 123},
            dims=("y", "x"))
        ancvars.append(sc[statusname])
    sc[dvname] = xr.DataArray(
        fake_alti,
        dims=("y", "x"),
        attrs={
            "area": fake_area,
            "scaled_FillValue": 123,
            "ancillary_variables": ancvars,
            "valid_range": rng})

    def _fake_get_varname(info, info_type="file_key"):
        return altname or dvname

    with mock.patch("satpy.readers.nwcsaf_nc.NcNWCSAF._get_varname_in_file") as srnN_:
        srnN_.side_effect = _fake_get_varname
        sc.load([comp])
    im = get_enhanced_image(sc[comp])
    if flavour == "geo":
        assert im.mode == "P"
        np.testing.assert_array_equal(im.data.coords["bands"], ["P"])
        if dtp == "float64":
            np.testing.assert_allclose(
                im.data.sel(bands="P"),
                ((fake_alti - rng[0]) * (255 / np.ptp(rng))).round())
        else:
            np.testing.assert_allclose(im.data.sel(bands="P"), fake_alti)


@pytest.mark.parametrize("name",
                         ["stretch",
                          "gamma",
                          "invert",
                          "piecewise_linear_stretch",
                          "cira_stretch",
                          "reinhard_to_srgb",
                          "btemp_threshold",
                         ]
                         )
def test_stretching_warns(name):
    """Test that there's a warning when importing stretching functions from old location."""
    from satpy import enhancements
    with pytest.warns(UserWarning, match="has been moved to"):
        getattr(enhancements, name)


def test_jma_true_color_repropdution_warns():
    """Test that there's a warning when importing jma_true_color_reproduction from old location."""
    with pytest.warns(UserWarning, match="has been moved to"):
        from satpy.enhancements import jma_true_color_reproduction  # noqa


def test_convolution_warns():
    """Test that there's a warning when importing three_d_effect from old location."""
    with pytest.warns(UserWarning, match="has been moved to"):
        from satpy.enhancements import three_d_effect  # noqa


@pytest.mark.parametrize("name",
                         ["exclude_alpha",
                          "on_separate_bands",
                          "using_map_blocks",
                         ]
                         )
def test_wrappers_warns(name):
    """Test that there's a warning when importing wrapper functions from old location."""
    from satpy import enhancements
    with pytest.warns(UserWarning, match="has been moved to"):
        getattr(enhancements, name)


@pytest.mark.parametrize("name",
                         ["lookup",
                          "colorize",
                          "palettize",
                          "create_colormap",
                         ]
                         )
def test_color_mapping_warns(name):
    """Test that there's a warning when importing color mapping functions from old location."""
    from satpy import enhancements
    with pytest.warns(UserWarning, match="has been moved to"):
        getattr(enhancements, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The hrit ahi reader tests package."""
from __future__ import annotations

import datetime as dt
from pathlib import Path
from unittest import mock

import numpy as np
import pytest

from satpy.readers.hrit_jma import (
    AREA_NAMES,
    FULL_DISK,
    HIMAWARI8,
    NORTH_HEMIS,
    PLATFORMS,
    SOUTH_HEMIS,
    UNKNOWN_AREA,
    UNKNOWN_PLATFORM,
    HRITJMAFileHandler,
)
from satpy.tests.utils import make_dataid


def create_fake_ahi_hrit(
        hrit_path: Path,
        num_rows: int = 11000,  # default 1km
        num_cols: int = 11000,  # default 1km
        is_vis: bool = True,
        annotation: str | None = None,
        metadata_overrides: dict | None = None,
) -> None:
    """Create a fake AHI HRIT file on disk."""
    header_data = _get_fake_header_data(num_rows, num_cols, is_vis, annotation or hrit_path.name)
    _update_header_with_metadata(header_data, metadata_overrides)
    # incremement by 1 but roll over to uint16 big-endian
    data_arr = np.arange(num_rows * num_cols, dtype=">u2")
    data_arr[-1] = 65535  # always make sure the last one is a fill value

    with hrit_path.open(mode="wb") as fp:
        for header_arr in header_data:
            header_arr.tofile(fp)
        data_arr.tofile(fp)


def _get_fake_header_data(
        num_rows: int,
        num_cols: int,
        is_vis: bool,
        annotation: str,
) -> tuple[np.ndarray, ...]:
    coff_loffs_bytes = _get_line_offsets(num_rows, num_cols)
    acq_times_bytes = _get_acq_time(num_rows)
    if is_vis:
        idf = "$HALFTONE:=16\r_NAME:=VISIBLE\r_UNIT:=ALBEDO(%)\r0:=-0.10\r1023:=100.00\r65535:=100.00\r"
    else:
        idf = "$HALFTONE:=16\r_NAME:=INFRARED\r_UNIT:=KELVIN\r0:=329.98\r1023:=130.02\r65535:=130.02\r"

    header_data = (
        # header 0
        np.void((0, 16), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        # total_header_length is updated below
        np.void((0, 0, 0),
                dtype=[("file_type", "u1"), ("total_header_length", ">u4"), ("data_field_length", ">u8")]),

        # header 1
        np.void((1, 9), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((16, num_cols, num_rows, 0),
                dtype=[("number_of_bits_per_pixel", "u1"), ("number_of_columns", ">u2"), ("number_of_lines", ">u2"),
                       ("compression_flag_for_data", "u1")]),

        # header 2
        np.void((2, 51), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((b"GEOS(140.70)                    ", 10233128, 10233128, 5500, 5500),
                dtype=[("projection_name", "S32"),
                       ("cfac", ">i4"), ("lfac", ">i4"),
                       ("coff", ">i4"), ("loff", ">i4")]),

        # header 3
        np.void((3, len(idf) + 3), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.bytes_(idf),

        # header 4
        np.void((4, len(annotation) + 3), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.bytes_(annotation),

        # header 5
        np.void((5, 10), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((64, (22167, 11162999)),
                dtype=[("cds_p_field", "u1"), ("timestamp", [("Days", ">u2"), ("Milliseconds", ">u4")])]),

        # header 128
        np.void((128, 7), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((0, 1, 1),
                dtype=[("image_segm_seq_no", "u1"), ("total_no_image_segm", "u1"), ("line_no_image_segm", ">u2")]),

        # header 130
        np.void((130, len(coff_loffs_bytes) + 3), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.bytes_(coff_loffs_bytes),

        # header 131
        np.void((131, len(acq_times_bytes) + 3), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.bytes_(acq_times_bytes),

        # header 132
        np.void((132, 12), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.array(b"NO_ERROR\r", dtype="|S9"),
    )
    total_header_length = sum([x.size * x.dtype.itemsize for x in header_data])
    header_data[1]["total_header_length"] = total_header_length
    return header_data


def _update_header_with_metadata(header_data: tuple[np.ndarray, ...], metadata_overrides: dict | None) -> None:
    if metadata_overrides is None:
        return
    for header_arr in header_data:
        keys_to_update = set(metadata_overrides.keys()) & set(header_arr.dtype.fields or [])
        for key in keys_to_update:
            new_val = metadata_overrides[key]
            if np.issubdtype(header_arr[key].dtype, np.bytes_):
                # null-pad the provided bytes
                new_val = np.array(new_val, dtype=header_arr[key].dtype)
            header_arr[key] = new_val


def _get_line_offsets(num_rows: int, num_cols: int) -> bytes:
    coff_loffs = f"LINE:=1\rCOFF:={num_cols / 2}\rLOFF:={num_rows / 2}"
    for lnum in range(1000, num_rows + 1, 1000):
        coff_loffs += f"LINE:={lnum}\rCOFF:={num_cols / 2}\rLOFF:={num_rows / 2}"
        if lnum != num_rows:
            coff_loffs += f"LINE:={lnum + 1}\rCOFF:={num_cols / 2}\rLOFF:={num_rows / 2}"
    return coff_loffs.encode()


def _get_acq_time(nlines):
    """Get sample header entry for scanline acquisition times.

    Lines: 1, 21, 41, 61, ..., nlines
    Times: 1970-01-01 00:00 + (1, 21, 41, 61, ..., nlines) seconds

    So the interpolated times are expected to be 1970-01-01 +
    (1, 2, 3, 4, ..., nlines) seconds. Note that there will be some
    floating point inaccuracies, because timestamps are stored
    with only 6 decimals precision.
    """
    # AHI:
    # 11000 == 1km
    # 2750 == 4km
    mjd_1970 = 40587.0
    lines_sparse = np.array(list(range(1, nlines, 20)) + [nlines])
    times_sparse = mjd_1970 + lines_sparse / 24 / 3600
    acq_time_s = ["LINE:={}\rTIME:={:.6f}\r".format(line, time)
                  for line, time in zip(lines_sparse, times_sparse)]
    acq_time_b = "".join(acq_time_s).encode()
    return acq_time_b


@pytest.mark.parametrize(
    ("channel", "is_vis", "shape"),
    [
        ("VIS", True, (11000, 11000)),
        ("IR4", False, (2750, 2750)),
    ],
)
def test_header_parsing(tmp_path, channel, is_vis, shape):
    """Test creating the file handler."""
    hrit_path = tmp_path / f"IMG_DK01{channel}_202509261940_001"
    create_fake_ahi_hrit(
        hrit_path,
        shape[0],
        shape[1],
        is_vis=is_vis,
    )
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})

    assert reader.mda["image_segm_seq_no"] == np.uint8(0)
    assert reader.mda["total_no_image_segm"] == np.uint8(1)
    assert reader.mda["projection_parameters"] == {
        "a": 6378169.00,
        "b": 6356583.80,
        "h": 35785831.00,
        "SSP_longitude": 140.7,
    }
    assert reader.mda["cfac"] == np.int32(10233128)
    assert reader.mda["lfac"] == np.int32(10233128)
    assert reader.mda["coff"] == np.int32(5500.0)
    assert reader.mda["loff"] == np.int32(5500.0)
    assert reader.mda["number_of_lines"] == np.uint16(shape[0])
    assert reader.mda["number_of_columns"] == np.uint16(shape[1])
    assert reader.mda["planned_end_segment_number"] == np.uint8(1)
    assert reader.mda["planned_start_segment_number"] == np.uint8(1)
    assert reader.mda["segment_sequence_number"] == np.uint8(0)
    assert reader.mda["unit"] == ("ALBEDO(%)" if is_vis else "KELVIN")
    assert reader.projection_name == "GEOS(140.70)"

    # Check calibration table
    if is_vis:
        cal_expected = np.array([[0, -0.1],
                                 [1023,  100],
                                 [65535,  100]])
    else:
        cal_expected = np.array([[0, 329.98],
                                 [1023, 130.02],
                                 [65535, 130.02]])
    assert np.all(reader.calibration_table == cal_expected)

    # Check if scanline timestamps are there (dedicated test below)
    assert isinstance(reader.acq_time, np.ndarray)
    assert reader.acq_time.dtype == np.dtype("datetime64[ns]")

    # Check platform
    assert reader.platform == HIMAWARI8
    hrit_path.unlink()


@pytest.mark.parametrize(
    ("segno", "is_segmented"),
    [
        (0, False),
        (1, True),
        (8, True),
    ]
)
def test_segmented_checks(tmp_path, segno, is_segmented):
    """Test segments are identified."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(
        hrit_path,
        1100,
        11000,
        metadata_overrides={"image_segm_seq_no": segno},
    )
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})
    assert reader.is_segmented == is_segmented


@pytest.mark.parametrize(
    ("extra_filename_info", "area_id"),
    [
        ({"area": FULL_DISK}, FULL_DISK),
        ({"area": 1234}, UNKNOWN_AREA),
        ({}, UNKNOWN_AREA)
    ]
)
def test_check_areas(tmp_path, extra_filename_info, area_id):
    """Test area names coming from the filename."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(hrit_path)
    filename_info = {"start_time": dt.datetime.now()}
    filename_info.update(extra_filename_info)
    reader = HRITJMAFileHandler(hrit_path, filename_info, {})
    assert reader.area_id == area_id
    hrit_path.unlink()


@pytest.mark.parametrize(("proj_name", "platform"), list(PLATFORMS.items()) + [("MERCATOR(0.0)", UNKNOWN_PLATFORM)])
def test_get_platform(tmp_path, proj_name, platform, caplog):
    """Test platform identification."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(hrit_path, metadata_overrides={"projection_name": proj_name})
    filename_info = {"start_time": dt.datetime.now()}
    reader = HRITJMAFileHandler(hrit_path, filename_info, {})

    reader.projection_name = proj_name
    assert reader._get_platform() == platform

    if proj_name == "invalid":
        assert "Unable to determine platform" in caplog.text
    hrit_path.unlink()


@pytest.mark.parametrize(
    ("mda_info", "area_name", "extent"),
    [
        # Non-segmented, full disk
        (
            {
                "loff": 1375.0, "coff": 1375.0,
                "nlines": 2750, "ncols": 2750,
                "segno": 0, "numseg": 1,
            },
            FULL_DISK,
            (-5498000.088960204, -5498000.088960204, 5502000.089024927, 5502000.089024927),
        ),
        # Non-segmented, northern hemisphere
        (
            {
                "loff": 1325.0, "coff": 1375.0,
                "nlines": 1375, "ncols": 2750,
                "segno": 0, "numseg": 1,
            },
            NORTH_HEMIS,
            (-5498000.088960204, -198000.00320373234, 5502000.089024927, 5302000.085788833),
        ),
        # Non-segmented, southern hemisphere
        (
            {
                "loff": 50, "coff": 1375.0,
                "nlines": 1375, "ncols": 2750,
                "segno": 0, "numseg": 1,
            },
            SOUTH_HEMIS,
            (-5498000.088960204, -5298000.085724112, 5502000.089024927, 202000.0032684542),
        ),
        # Segmented, segment #1
        (
            {
                "loff": 1375.0, "coff": 1375.0,
                "nlines": 275, "ncols": 2750,
                "segno": 1, "numseg": 10,
            },
            FULL_DISK,
            (-5498000.088960204, 4402000.071226413, 5502000.089024927, 5502000.089024927),
        ),
        # Segmented, segment #7
        (
            {
                "loff": 1375.0, "coff": 1375.0,
                "nlines": 275, "ncols": 2750,
                "segno": 7, "numseg": 10,
            },
            FULL_DISK,
            (-5498000.088960204, -2198000.035564665, 5502000.089024927, -1098000.0177661523),
        ),
    ]
)
def test_get_area_def(tmp_path, mda_info, area_name, extent):
    """Test getting an AreaDefinition."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(
        hrit_path,
        mda_info["nlines"],
        mda_info["ncols"],
        metadata_overrides={
            "loff": mda_info["loff"],
            "coff": mda_info["coff"],
            "image_segm_seq_no": mda_info["segno"],
            "total_no_image_segm": mda_info["numseg"],
        },
    )
    reader = HRITJMAFileHandler(
        hrit_path,
        {"start_time": dt.datetime.now(), "area": area_name},
        {})

    area = reader.get_area_def("some_id")
    assert area.area_extent == extent
    assert area.description == AREA_NAMES[area_name]["long"]


@pytest.mark.parametrize("calibration", ["counts", "reflectance", "brightness_temperature"])
def test_calibrate(tmp_path, calibration):
    """Test calibration."""
    num_rows = 275
    num_cols = 1375
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    units = {
        "counts": "1",
        "reflectance": "%",
        "brightness_temperature": "K",
    }[calibration]
    create_fake_ahi_hrit(
        hrit_path,
        num_rows,
        num_cols,
        is_vis=calibration == "reflectance",
        # Choose an area near the subsatellite point to avoid masking of space pixels
        metadata_overrides={
            "loff": 1375.0,
            "coff": 1375.0,
            "image_segm_seq_no": 1,
            "total_no_image_segm": 10,
        },
    )
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})
    key = make_dataid(name="VIS", calibration=calibration)
    res = reader.get_dataset(
        key,
        {
            "units": units,
            "sensor": "ahi",
        },
    )

    exp_counts = np.arange(49658, 49658 + 200, 1, dtype=">u2").astype(np.float32)
    exp = {
        "counts": exp_counts,
        "reflectance": np.interp(exp_counts, [0, 1023, 65535], [-0.10, 100.0, 100.0]),
        "brightness_temperature": np.interp(exp_counts, [0, 1023, 65535], [329.98, 130.02, 130.02]),
    }
    res_np = res.data.compute()
    assert res_np.dtype == res.dtype
    assert res_np.dtype == np.float32 if calibration == "counts" else np.float64
    np.testing.assert_allclose(res_np[-1, 588:788], exp[calibration])


def test_mask_space(tmp_path):
    """Test masking of space pixels."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(
        hrit_path,
        275,
        1375,
        # Choose an area near the subsatellite point to avoid masking of space pixels
        metadata_overrides={
            "loff": 1375.0,
            "coff": 1375.0,
            "image_segm_seq_no": 1,
            "total_no_image_segm": 10,
        },
    )
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})

    key = make_dataid(name="VIS", calibration="counts")
    res = reader.get_dataset(key, {"units": "1", "sensor": "ahi"})
    res_np = res.data.compute()
    assert res_np.dtype == res.dtype
    assert res_np.dtype == np.float32

    # First line of the segment should be space, in the middle of the
    # last line there should be some valid pixels
    np.testing.assert_allclose(res_np[0, :], np.nan)
    exp = np.arange(49658, 49658 + 200, 1, dtype=">u2").astype(np.float32)
    np.testing.assert_array_equal(res_np[-1, 588:788], exp)


def test_get_dataset(tmp_path):
    """Test getting a dataset."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(
        hrit_path,
        275,
        1375,
        # Choose an area near the subsatellite point to avoid masking of space pixels
        metadata_overrides={
            "loff": 1375.0,
            "coff": 1375.0,
            "image_segm_seq_no": 1,
            "total_no_image_segm": 10,
        },
    )
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})

    key = make_dataid(name="VIS", calibration="reflectance")
    with mock.patch.object(reader, "_mask_space", wraps=reader._mask_space) as mask_space, \
            mock.patch.object(reader, "calibrate", wraps=reader.calibrate) as calibrate:
        res = reader.get_dataset(key, {"units": "%", "sensor": "ahi"})
        mask_space.assert_called()
        calibrate.assert_called()

    # Check attributes
    assert res.attrs["units"] == "%"
    assert res.attrs["sensor"] == "ahi"
    assert res.attrs["platform_name"] == HIMAWARI8
    assert res.attrs["orbital_parameters"] == {"projection_longitude": 140.7,
                                               "projection_latitude": 0.0,
                                               "projection_altitude": 35785831.0}

    # Check if acquisition time is a coordinate
    assert "acq_time" in res.coords


def test_sensor_mismatch(tmp_path, caplog):
    """Test that a file for a different sensor is detected."""
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(hrit_path)
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})

    key = make_dataid(name="VIS", calibration="reflectance")
    reader.get_dataset(key, {"units": "%", "sensor": "jami"})
    assert "Sensor-Platform mismatch" in caplog.text
    hrit_path.unlink()


@pytest.mark.parametrize("projection_name", ["GEOS(140.70)", "GEOS(145.00)"])
def test_get_acq_time(tmp_path, projection_name):
    """Test computation of scanline acquisition times."""
    dt_line = np.arange(1, 11000+1).astype("timedelta64[s]")
    acq_time_exp = np.datetime64("1970-01-01", "ns") + dt_line
    # Results are not exactly identical because timestamps are stored in
    # the header with only 6 decimals precision (max diff here: 45 msec).

    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(hrit_path, metadata_overrides={"projection_name": projection_name})
    reader = HRITJMAFileHandler(hrit_path, {"start_time": dt.datetime.now()}, {})
    np.testing.assert_allclose(reader.acq_time.astype(np.int64),
                               acq_time_exp.astype(np.int64),
                               atol=45000000)
    hrit_path.unlink()


@pytest.mark.parametrize("projection_name", ["GEOS(140.70)", "GEOS(145.00)"])
@pytest.mark.parametrize("use_acq_time", [None, False, True])
def test_start_time_from_aqc_time(tmp_path, projection_name, use_acq_time):
    """Test that by the datetime from the metadata returned when `use_acquisition_time_as_start_time=True`."""
    start_time = dt.datetime(2022, 1, 20, 12, 10)
    hrit_path = tmp_path / "IMG_DK01VIS_202509261940_001"
    create_fake_ahi_hrit(hrit_path, metadata_overrides={"projection_name": projection_name})
    reader_kwargs = {"use_acquisition_time_as_start_time": True} if use_acq_time else {}
    reader = HRITJMAFileHandler(hrit_path, {"start_time": start_time}, {}, **reader_kwargs)

    if use_acq_time:
        assert reader.start_time == dt.datetime(1970, 1, 1, 0, 0, 1, 36799)
        assert reader.end_time == dt.datetime(1970, 1, 1, 3, 3, 20, 16000)
    else:
        assert reader.start_time == start_time
    hrit_path.unlink()


@pytest.mark.parametrize(
    ("mjd", "dt_str"),
    [
        (0, "1858-11-17"),
        (40587.5, "1970-01-01 12:00"),
    ]
)
def test_mjd2datetime64(mjd, dt_str):
    """Test conversion from modified julian day to datetime64."""
    from satpy.readers.hrit_jma import mjd2datetime64
    assert mjd2datetime64(np.array([mjd])) == np.datetime64(dt_str, "ns")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Satpy Configuration directory and file handling."""
from __future__ import annotations

import ast
import glob
import logging
import os
import sys
import tempfile
from collections import OrderedDict
from importlib.metadata import EntryPoint, entry_points
from importlib.resources import files as impr_files
from typing import Iterable

from donfig import Config
from platformdirs import AppDirs

from satpy._compat import cache

LOG = logging.getLogger(__name__)

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
# FIXME: Use package_resources?
PACKAGE_CONFIG_PATH = os.path.join(BASE_PATH, "etc")

_satpy_dirs = AppDirs(appname="satpy", appauthor="pytroll")
_CONFIG_DEFAULTS = {
    "tmp_dir": tempfile.gettempdir(),
    "cache_dir": _satpy_dirs.user_cache_dir,
    "cache_lonlats": False,
    "cache_sensor_angles": False,
    "config_path": [],
    "data_dir": _satpy_dirs.user_data_dir,
    "demo_data_dir": ".",
    "download_aux": True,
    "sensor_angles_position_preference": "actual",
    "readers": {
        "clip_negative_radiances": False,
    },
}

# Satpy main configuration object
# See https://donfig.readthedocs.io/en/latest/configuration.html
# for more information.
#
# Configuration values will be loaded from files at:
# 1. The builtin package satpy.yaml (not present currently)
# 2. $SATPY_ROOT_CONFIG (default: /etc/satpy/satpy.yaml)
# 3. <python-env-prefix>/etc/satpy/satpy.yaml
# 4. ~/.config/satpy/satpy.yaml
# 5. ~/.satpy/satpy.yaml
# 6. $SATPY_CONFIG_PATH/satpy.yaml if present (colon separated)
_CONFIG_PATHS = [
    os.path.join(PACKAGE_CONFIG_PATH, "satpy.yaml"),
    os.getenv("SATPY_ROOT_CONFIG", os.path.join("/etc", "satpy", "satpy.yaml")),
    os.path.join(sys.prefix, "etc", "satpy", "satpy.yaml"),
    os.path.join(_satpy_dirs.user_config_dir, "satpy.yaml"),
    os.path.join(os.path.expanduser("~"), ".satpy", "satpy.yaml"),
]
# The above files can also be directories. If directories all files
# with `.yaml`., `.yml`, or `.json` extensions will be used.

_ppp_config_dir = os.getenv("PPP_CONFIG_DIR", None)
_satpy_config_path = os.getenv("SATPY_CONFIG_PATH", None)

if _ppp_config_dir is not None and _satpy_config_path is None:
    LOG.warning("'PPP_CONFIG_DIR' is deprecated. Please use 'SATPY_CONFIG_PATH' instead.")
    _satpy_config_path = _ppp_config_dir

if _satpy_config_path is not None:
    if _satpy_config_path.startswith("["):
        # 'SATPY_CONFIG_PATH' is set by previous satpy config as a reprsentation of a 'list'
        # need to use 'ast.literal_eval' to parse the string back to a list
        _satpy_config_path_list = ast.literal_eval(_satpy_config_path)
    else:
        # colon-separated are ordered by custom -> builtins
        # i.e. last-applied/highest priority to first-applied/lowest priority
        _satpy_config_path_list = _satpy_config_path.split(os.pathsep)

    os.environ["SATPY_CONFIG_PATH"] = repr(_satpy_config_path_list)
    for config_dir in _satpy_config_path_list:
        _CONFIG_PATHS.append(os.path.join(config_dir, "satpy.yaml"))

_ancpath = os.getenv("SATPY_ANCPATH", None)
_data_dir = os.getenv("SATPY_DATA_DIR", None)
if _ancpath is not None and _data_dir is None:
    LOG.warning("'SATPY_ANCPATH' is deprecated. Please use 'SATPY_DATA_DIR' instead.")
    os.environ["SATPY_DATA_DIR"] = _ancpath

config = Config("satpy", defaults=[_CONFIG_DEFAULTS], paths=_CONFIG_PATHS)


def get_config_path_safe():
    """Get 'config_path' and check for proper 'list' type."""
    config_path = config.get("config_path")
    if not isinstance(config_path, list):
        raise ValueError("Satpy config option 'config_path' must be a "
                         "list, not '{}'".format(type(config_path)))
    return config_path


def get_entry_points_config_dirs(group_name: str, include_config_path: bool = True) -> list[str]:
    """Get the config directories for all entry points of given name."""
    dirs: list[str] = []
    for entry_point in cached_entry_point(group_name):
        module = _entry_point_module(entry_point)
        new_dir = str(impr_files(module) / "etc")
        if not dirs or dirs[-1] != new_dir:
            dirs.append(new_dir)
    if include_config_path:
        dirs.extend(config.get("config_path")[::-1])
    return dirs


@cache
def cached_entry_point(group_name: str) -> Iterable[EntryPoint]:
    """Return entry_point for specified ``group``.

    This is a dummy proxy to allow caching and provide compatibility between
    versions of Python and importlib_metadata.

    """
    try:
        # mypy in pre-commit currently checks for Python 3.8 compatibility
        # this line is for Python 3.10+ so it will fail checks
        return entry_points(group=group_name)  # type: ignore
    except TypeError:
        # Python <3.10
        entry_points_list = entry_points()
        return entry_points_list.get(group_name, [])


def _entry_point_module(entry_point):
    try:
        return entry_point.module
    except AttributeError:
        # Python 3.8
        return entry_point.value.split(":")[0].strip()


def config_search_paths(filename, search_dirs=None, **kwargs):
    """Get series of configuration base paths where Satpy configs are located."""
    if search_dirs is None:
        search_dirs = get_config_path_safe()[::-1]

    paths = [filename, os.path.basename(filename)]
    paths += [os.path.join(search_dir, filename) for search_dir in search_dirs]
    paths += [os.path.join(PACKAGE_CONFIG_PATH, filename)]
    paths = [os.path.abspath(path) for path in paths]

    if kwargs.get("check_exists", True):
        paths = [x for x in paths if os.path.isfile(x)]

    paths = list(OrderedDict.fromkeys(paths))
    # flip the order of the list so builtins are loaded first
    return paths[::-1]


def glob_config(pattern, search_dirs=None):
    """Return glob results for all possible configuration locations.

    Note: This method does not check the configuration "base" directory if the pattern includes a subdirectory.
          This is done for performance since this is usually used to find *all* configs for a certain component.
    """
    patterns = config_search_paths(pattern, search_dirs=search_dirs,
                                   check_exists=False)
    for pattern_fn in patterns:
        for path in glob.iglob(pattern_fn):
            yield path


def get_config_path(filename):
    """Get the path to the highest priority version of a config file."""
    paths = config_search_paths(filename)
    for path in paths[::-1]:
        if os.path.exists(path):
            return path
    raise FileNotFoundError("Could not find file in configuration path: "
                            "'{}'".format(filename))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Base reader for the L1 HDF data from the AGRI and GHI instruments aboard the FengYun-4A/B satellites.

The files read by this reader are described in the official Real Time Data Service:

    http://fy4.nsmc.org.cn/data/en/data/realtime.html

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import numpy.typing as npt
import xarray as xr

from satpy._compat import cached_property
from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.hdf5 import HDF5FileHandler

logger = logging.getLogger(__name__)

RESOLUTION_LIST = [250, 500, 1000, 2000, 4000]


class FY4Base(HDF5FileHandler):
    """The base class for the FengYun4 AGRI and GHI readers."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init filehandler."""
        super(FY4Base, self).__init__(filename, filename_info, filetype_info)

        self.sensor = filename_info["instrument"]

        # info of 250m, 500m, 1km, 2km and 4km data
        self._COFF_list = [21983.5, 10991.5, 5495.5, 2747.5, 1373.5]
        self._LOFF_list = [21983.5, 10991.5, 5495.5, 2747.5, 1373.5]

        self._CFAC_list = [163730199.0, 81865099.0, 40932549.0, 20466274.0, 10233137.0]
        self._LFAC_list = [163730199.0, 81865099.0, 40932549.0, 20466274.0, 10233137.0]

        self.PLATFORM_NAMES = {"FY4A": "FY-4A",
                               "FY4B": "FY-4B",
                               "FY4C": "FY-4C"}

        try:
            self.PLATFORM_ID = self.PLATFORM_NAMES[filename_info["platform_id"]]
        except KeyError:
            raise KeyError(f"Unsupported platform ID: {filename_info['platform_id']}")
        self.CHANS_ID = "NOMChannel"
        self.SAT_ID = "NOMSatellite"
        self.SUN_ID = "NOMSun"

    @staticmethod
    def scale(dn, slope, offset):
        """Convert digital number (DN) to calibrated quantity through scaling.

        Args:
            dn: Raw detector digital number
            slope: Slope
            offset: Offset

        Returns:
            Scaled data

        """
        ref = dn * slope + offset
        ref = ref.clip(min=0)
        ref.attrs = dn.attrs

        return ref

    def _apply_lut(self, data: xr.DataArray, lut: npt.NDArray[np.float32]) -> xr.DataArray:
        """Calibrate digital number (DN) by applying a LUT.

        Args:
            data: Raw detector digital number
            lut: the look up table
        Returns:
            Calibrated quantity
        """
        # append nan to the end of lut for fillvalue
        fill_value = data.attrs.get("FillValue")
        if fill_value is not None and fill_value.item() <= lut.shape[0] - 1:
            # If LUT includes the fill_value, remove that entry and everything
            # after it.
            # Ex. C07 has a LUT of 65536 elements, but fill value is 65535
            # This is considered a bug in the input file format
            lut = lut[:fill_value.item()]

        lut = np.append(lut, np.nan)
        data.data = da.where(data.data >= lut.shape[0], lut.shape[0] - 1, data.data)
        res = data.data.map_blocks(self._getitem, lut, dtype=lut.dtype)
        res = xr.DataArray(res, dims=data.dims,
                           attrs=data.attrs, coords=data.coords)

        return res

    @staticmethod
    def _getitem(block, lut):
        return lut[block]

    @cached_property
    def reflectance_coeffs(self):
        """Retrieve the reflectance calibration coefficients from the HDF file."""
        # using the corresponding SCALE and OFFSET
        if self.PLATFORM_ID == "FY-4A":
            cal_coef = "CALIBRATION_COEF(SCALE+OFFSET)"
        elif self.PLATFORM_ID == "FY-4B":
            cal_coef = "Calibration/CALIBRATION_COEF(SCALE+OFFSET)"
        else:
            raise KeyError(f"Unsupported platform ID for calibration: {self.PLATFORM_ID}")
        return self.get(cal_coef).values

    def calibrate(self, data, ds_info, ds_name, file_key):
        """Calibrate the data."""
        # Check if calibration is present, if not assume dataset is an angle
        calibration = ds_info.get("calibration")
        # Return raw data in case of counts or no calibration
        if calibration in ("counts", None):
            data.attrs["units"] = ds_info["units"]
            ds_info["valid_range"] = data.attrs["valid_range"]
            ds_info["fill_value"] = data.attrs["FillValue"].item()
        elif calibration == "reflectance":
            channel_index = int(file_key[-2:]) - 1
            data = self.calibrate_to_reflectance(data, channel_index, ds_info)
        elif calibration == "brightness_temperature":
            data = self.calibrate_to_bt(data, ds_info, ds_name)
        elif calibration == "radiance":
            raise NotImplementedError("Calibration to radiance is not supported.")
        # Apply range limits, but not for counts or we convert to float!
        if calibration != "counts":
            data = data.where((data >= min(ds_info["valid_range"])) &
                              (data <= max(ds_info["valid_range"])))
        else:
            data.attrs["_FillValue"] = data.attrs["FillValue"].item()
        return data

    def calibrate_to_reflectance(self, data, channel_index, ds_info):
        """Calibrate to reflectance [%]."""
        logger.debug("Calibrating to reflectances")
        # using the corresponding SCALE and OFFSET
        if self.sensor != "AGRI" and self.sensor != "GHI":
            raise ValueError(f"Unsupported sensor type: {self.sensor}")

        coeffs = self.reflectance_coeffs
        num_channel = coeffs.shape[0]

        if self.sensor == "AGRI" and num_channel == 1:
            # only channel_2, resolution = 500 m
            channel_index = 0
        data.data = da.where(data.data == data.attrs["FillValue"].item(), np.nan, data.data)
        data.attrs["scale_factor"] = coeffs[channel_index, 0].item()
        data.attrs["add_offset"] = coeffs[channel_index, 1].item()
        data = self.scale(data, data.attrs["scale_factor"], data.attrs["add_offset"])
        data *= 100
        ds_info["valid_range"] = (data.attrs["valid_range"] * data.attrs["scale_factor"] + data.attrs["add_offset"])
        ds_info["valid_range"] = ds_info["valid_range"] * 100
        return data

    def calibrate_to_bt(self, data, ds_info, ds_name):
        """Calibrate to Brightness Temperatures [K]."""
        logger.debug("Calibrating to brightness_temperature")

        if self.sensor not in ["GHI", "AGRI"]:
            raise ValueError("Error, sensor must be GHI or AGRI.")

        # The key is sometimes prefixes with `Calibration/` so we try both options here
        lut_key = ds_info.get("lut_key", ds_name)
        try:
            lut = self[lut_key]
        except KeyError:
            lut_key = f'Calibration/{ds_info.get("lut_key", ds_name)}'
            lut = self[lut_key]

        # the value of dn is the index of brightness_temperature
        data = self._apply_lut(data, lut.compute().data)
        ds_info["valid_range"] = lut.attrs["valid_range"]
        return data

    @property
    def start_time(self):
        """Get the start time."""
        start_time = self["/attr/Observing Beginning Date"] + "T" + self["/attr/Observing Beginning Time"] + "Z"
        try:
            return dt.datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            # For some data there is no sub-second component
            return dt.datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%SZ")

    @property
    def end_time(self):
        """Get the end time."""
        end_time = self["/attr/Observing Ending Date"] + "T" + self["/attr/Observing Ending Time"] + "Z"
        try:
            return dt.datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            # For some data there is no sub-second component
            return dt.datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%SZ")

    def get_area_def(self, key):
        """Get the area definition."""
        # Coordination Group for Meteorological Satellites LRIT/HRIT Global Specification
        # https://www.cgms-info.org/documents/cgms-lrit-hrit-global-specification-(v2-8-of-30-oct-2013).pdf
        res = key["resolution"]
        pdict = {}

        begin_cols = float(self.file_content["/attr/Begin Pixel Number"])
        end_lines = float(self.file_content["/attr/End Line Number"])
        pdict["coff"] = self._COFF_list[RESOLUTION_LIST.index(res)] - begin_cols + 1
        pdict["loff"] = -self._LOFF_list[RESOLUTION_LIST.index(res)] + end_lines + 1

        pdict["cfac"] = self._CFAC_list[RESOLUTION_LIST.index(res)]
        pdict["lfac"] = self._LFAC_list[RESOLUTION_LIST.index(res)]
        try:
            pdict["a"] = float(self.file_content["/attr/Semimajor axis of ellipsoid"])
        except KeyError:
            pdict["a"] = float(self.file_content["/attr/dEA"])
        if pdict["a"] < 10000:
            pdict["a"] = pdict["a"] * 1E3  # equator radius (m)
        try:
            pdict["b"] = float(self.file_content["/attr/Semiminor axis of ellipsoid"])
        except KeyError:
            pdict["b"] = pdict["a"] * (1 - 1 / self.file_content["/attr/dObRecFlat"])  # polar radius (m)

        pdict["h"] = self.file_content["/attr/NOMSatHeight"]  # the altitude of satellite (m)
        if pdict["h"] > 42000000.0:
            pdict["h"] = pdict["h"] - pdict["a"]

        pdict["ssp_lon"] = float(self.file_content["/attr/NOMCenterLon"])
        pdict["nlines"] = float(self.file_content["/attr/RegLength"])
        pdict["ncols"] = float(self.file_content["/attr/RegWidth"])

        pdict["scandir"] = "N2S"
        pdict["a_desc"] = "FY-4 {} area".format(self.filename_info["observation_type"])
        pdict["a_name"] = f'{self.filename_info["observation_type"]}_{res}m'
        pdict["p_id"] = f"FY-4, {res}m"

        area_extent = get_area_extent(pdict)
        area_extent = (area_extent[0],
                       area_extent[1],
                       area_extent[2],
                       area_extent[3])

        area = get_area_definition(pdict, area_extent)

        return area



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Common utility modules used for LI mock-oriented unit tests."""

import datetime as dt

import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import RANDOM_GEN

# mapping of netcdf type code to numpy data type:
TYPE_MAP = {
    "i1": np.int8,
    "i2": np.int16,
    "i4": np.int32,
    "i8": np.int64,
    "u1": np.uint8,
    "u2": np.uint16,
    "u4": np.uint32,
    "u8": np.uint64,
    "f4": np.float32,
    "f8": np.float64,
}

def rand_type(num, dtype):
        return RANDOM_GEN.integers(low=np.iinfo(dtype).min, high=np.iinfo(dtype).max - 1, size=num, dtype=dtype)

def l2_le_schema(settings=None):
    """Define schema for LI L2 LE product."""
    settings = settings or {}
    nobs = settings.get("num_obs", 123)
    nchunks = settings.get("num_chunks", 23)
    nfilters = settings.get("num_filters", 2)

    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", "data/"),
        "dimensions": {
            "unfiltered_events": nobs,
            "l1b_chunks": nchunks,
            "l1b_offsets": nchunks,
            "filters": nfilters,
            "scalar": 1,
        },
        "variables": {},
        "sector_variables": {
            "event_id": {
                "format": "u4",
                "shape": ("unfiltered_events",),
                "long_name": "ID of LI L2 Event",
                "default_data": lambda: rand_type(nobs, np.uint32)
            },
            "group_id": {
                "format": "u4",
                "shape": ("unfiltered_events",),
                "long_name": "ID of associated LI L2 Group object",
                "default_data": lambda: rand_type(nobs, np.uint32)
            },
            "l1b_chunk_ids": {
                "format": "u4",
                "shape": ("l1b_chunks",),
                "long_name": "Array of L1b event chunk IDs",
                "default_data": lambda: rand_type(nchunks, np.uint32)
            },
            "l1b_chunk_offsets": {
                "format": "u4",
                "shape": ("l1b_offsets",),
                "long_name": "Array offset for L1b event chunk boundaries",
                "default_data": lambda: np.arange(nchunks)
            },
            "l1b_window": {
                "format": "u4",
                "shape": ("unfiltered_events",),
                "long_name": "window index of associated L1b event",
                "default_data": lambda: rand_type(nobs, np.uint32)
            },
            "filter_values": {
                "format": "u1",
                "shape": ("unfiltered_events", "filters",),
                "fill_value": 255,
                "scale_factor": 0.004,
                "add_offset": 0.0,
                "long_name": "L2 filter results",
                "default_data": lambda: rand_type((nobs, nfilters), np.uint8)
            },
            "epoch_time": {
                "format": "f8",
                "shape": ("scalar",),
                "long_name": "Start time of integration frame",
                "default_data": lambda: 1.234,
                "precision": "1 millisecond",
                "time_standard": "UTC",
                "standard_name": "time",
                "units": "seconds since 2000-01-01 00:00:00.0",
            },
            "time_offset": {
                "format": "f4",
                "shape": ("unfiltered_events",),
                "fill_value": 9.96921e36,
                "long_name": "Time offset from epoch time",
                "default_data": lambda: np.linspace(0.0, 1000.0, nobs),
                "units": "seconds",
            },
            "detector": {
                "format": "u1",
                "fill_value": 255,
                "shape": ("scalar",),
                "long_name": "ID of detector for this group",
                "default_data": lambda: 1,
                "meanings": "1 = detector_1, 2 = detector_2, 3 = detector_3, 4 = detector_4",
            },
            "l1b_filter_qa": {
                "format": "u1",
                "fill_value": 255,
                "shape": ("unfiltered_events",),
                "add_offset" : 0.0,
                "scale_offset": 0.004,
                "long_name": "L1b event confidence",
                "default_data": lambda : rand_type(nobs, np.uint8),
            },
            "l2_group_filter_qa": {
                "format": "u1",
                "fill_value": 255,
                "shape": ("unfiltered_events",),
                "add_offset" : 0.0,
                "scale_offset": 0.004,
                "long_name": "L2 group confidence",
                "default_data": lambda: (np.arange(nobs) + 10000),
            },
        }
    }


def l2_lef_schema(settings=None):
    """Define schema for LI L2 LEF product."""
    epoch_ts = dt.datetime(2000, 1, 1, 0, 0, 0, 0)
    start_time = dt.datetime.now()
    start_ts = (start_time - epoch_ts).total_seconds()

    settings = settings or {}

    nobs = settings.get("num_obs", 123)

    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", "data/"),
        "dimensions": {
            "events": nobs,
            "scalar": 1,
        },
        "variables": {
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ("scalar",),  # test explicitly the scalar case
                "long_name": "L1b event geolocation warning",
                "default_data": lambda: 0
            },
            "l1b_missing_warning": {
                "format": "i1",
                "shape": ("scalar",),
                "long_name": "Expected L1b inputs missing",
                "default_data": lambda: 0
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ("scalar",),
                "long_name": "L1b event radiometric warning",
                "default_data": lambda: 0
            },
        },
        "sector_variables": {
            "event_id": {
                "format": "u4",
                "shape": ("events",),
                "long_name": "ID of LI L2 Event",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "group_id": {
                "format": "u4",
                "shape": ("events",),
                "long_name": "ID of associated LI L2 Group object",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "flash_id": {
                "format": "u4",
                "shape": ("events",),
                "long_name": "ID of associated LI L2 Flash object",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "detector": {
                "format": "u1",
                "shape": ("scalar",),
                "fill_value": 255,
                "long_name": "ID of detector for this group",
                "meaning": "1 = detector_1, 2 = detector_2, 3 = detector_3, 4 = detector_4",
                "default_data": lambda: 1
            },
            "latitude": {
                "format": "i2",
                "shape": ("events",),
                "fill_value": -32767,
                "long_name": "Latitude of group",
                "units": "degrees_north",
                "standard_name": "latitude",
                "default_data": lambda: np.linspace(-90, 90, nobs)
            },
            "longitude": {
                "format": "i2",
                "shape": ("events",),
                "fill_value": -32767,
                "long_name": "Longitude of group",
                "units": "degrees_east",
                "standard_name": "longitude",
                "default_data": lambda: np.linspace(-180, 80, nobs)
            },
            "radiance": {
                "format": "u2",
                "shape": ("events",),
                "long_name": "Radiance of Flash",
                "standard_name": "radiance",
                "units": "mW.m-2.sr-1",
                "default_data": lambda: np.clip(np.round(RANDOM_GEN.normal(500, 100, nobs)), 1, 2 ** 16 - 1)
            },
            "event_filter_qa": {
                "format": "u2",
                "shape": ("events",),
                "fill_value": 255,
                "long_name": "L2 event pre-filtering quality assurance value",
                "default_data": lambda: RANDOM_GEN.integers(1, 2 ** 8 - 1, nobs)
            },
            "epoch_time": {
                "format": "f8",
                "shape": ("scalar",),
                "long_name": "Start time of integration frame",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "default_data": lambda: start_ts
            },
            "time_offset": {
                "format": "f4",
                "shape": ("events",),
                "long_name": "Time offset from epoch time",
                "units": "seconds",
                "default_data": lambda: RANDOM_GEN.uniform(1, 2 ** 31 - 1, nobs)
            },
            "detector_row": {
                "format": "u2",
                "shape": ("events",),
                "fill_value": 65535,
                "long_name": "Detector row position of event pixel",
                "units": "1",
                "default_data": lambda: RANDOM_GEN.integers(1, 1000, nobs)
            },
            "detector_column": {
                "format": "u2",
                "shape": ("events",),
                "fill_value": 65535,
                "long_name": "Detector column position of event pixel",
                "units": "1",
                "default_data": lambda: RANDOM_GEN.integers(1, 1000, nobs)
            },
        }
    }


def l2_lgr_schema(settings=None):
    """Define schema for LI L2 LGR product."""
    settings = settings or {}
    ngrps = settings.get("num_groups", 120)

    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", ""),
        "dimensions": {
            "groups": ngrps,
            "scalar": 1,
        },
        "variables": {
            "group_time": {
                "format": "f8",
                "shape": ("groups",),
                "long_name": "Start time of integration frame",
                "standard_name": "time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "precision": "0.001",
                "time_standard": "UTC",
                "default_data": lambda: np.linspace(-90, 90, ngrps)
            },
            "latitude": {
                "format": "i2",
                "shape": ("groups",),
                "long_name": "Latitude of group",
                "units": "degrees_north",
                "fill_value": -32767,
                "default_data": lambda: np.linspace(-90, 90, ngrps)
            },
            "longitude": {
                "format": "i2",
                "shape": ("groups",),
                "long_name": "Longitude of group",
                "fill_value": -32767,
                "units": "degrees_east",
                "default_data": lambda: np.linspace(-180, 80, ngrps)
            },
            "radiance": {
                "format": "u2",
                "shape": ("groups",),
                "long_name": "Radiance of group",
                "fill_value": 65535,
                "units": "mW.m-2.sr-1",
                "scale_factor": 0.5,
                "add_offset": 0.0,
                "default_data": lambda: rand_type(ngrps, np.uint16)
            },
           "group_id": {
                "format": "u4",
                "shape": ("groups",),
                "long_name": "LI L2 group IDs",
                "default_data": lambda: np.linspace(-180, 80, ngrps)
            },
           "flash_id": {
                "format": "u4",
                "shape": ("groups",),
                "long_name": "ID of associated LI L2 Flash object with each group",
                "default_data": lambda: np.linspace(-180, 80, ngrps)
            },
           "number_of_events": {
                "format": "u2",
                "shape": ("groups",),
                "long_name": "Number of events in each group",
                "default_data": lambda: np.linspace(-180, 80, ngrps)
            },
           "group_filter_qa": {
                "format": "u1",
                "shape": ("groups",),
                "fill_value": 255,
                "long_name": "L2 filtered group quality assurance value",
                "add_offset": 0.0,
                "scale_factor": 0.004,
                "default_data": lambda: np.linspace(-180, 80, ngrps)
            },
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ("scalar",),  # test explicitly the scalar case
                "long_name": "L1b event geolocation warning",
                "default_data": lambda: 0
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ("scalar",),
                "long_name": "L1b event radiometric warning",
                "default_data": lambda: 0
            },
        }
    }


def l2_lfl_schema(settings=None):
    """Define schema for LI L2 LFL product."""
    settings = settings or {}

    nobs = settings.get("num_obs", 1234)
    epoch = dt.datetime(2000, 1, 1)
    stime = (dt.datetime(2019, 1, 1) - epoch).total_seconds()
    etime = (dt.datetime(2019, 1, 2) - epoch).total_seconds()

    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", ""),
        "dimensions": {
            "flashes": nobs,
            "scalar": 1,
        },
        "variables": {
            "latitude": {
                "format": "i2",
                "shape": ("flashes",),
                "long_name": "Latitude of Flash",
                "standard_name": "latitude",
                "units": "degrees_north",
                "fill_value": -32767,
                "add_offset": 0.0,
                "scale_factor": 0.0027,
                # Note: using a default range of [-88.3deg, 88.3deg] to stay in
                # the available type range [-32727,32727] with scaling:
                "default_data": lambda: np.linspace(-88.3 / 0.0027, 88.3 / 0.0027, nobs)
            },
            "longitude": {
                "format": "i2",
                "shape": ("flashes",),
                "long_name": "Longitude of Flash",
                "standard_name": "longitude",
                "units": "degrees_east",
                "fill_value": -32767,
                "add_offset": 0.0,
                "scale_factor": 0.0027,
                # Note: using a default range of [-88.3deg, 88.3deg] to stay in
                # the available type range [-32727,32727] with scaling:
                "default_data": lambda: np.linspace(-88.3 / 0.0027, 88.3 / 0.0027, nobs)
            },
            "radiance": {
                "format": "u2",
                "shape": ("flashes",),
                "fill_value" : 65535,
                "long_name": "Radiance of Flash",
                "standard_name": "radiance",
                "units": "mW.m-2.sr-1",
                "default_data": lambda: np.round(RANDOM_GEN.normal(500, 100, nobs))
            },
            "flash_duration": {
                "format": "u2",
                "shape": ("flashes",),
                "long_name": "Flash duration",
                "standard_name": "flash_duration",
                "units": "ms",
                "default_data": lambda: np.linspace(0, 1000, nobs)
            },
            "flash_filter_confidence": {
                "format": "u1",
                "shape": ("flashes",),
                "fill_value": 255,
                "long_name": "L2 filtered flash confidence",
                "standard_name": "flash_filter_confidence",
                "default_data": lambda: np.clip(np.round(RANDOM_GEN.normal(20, 10, nobs)), 1, 2 ** 7 - 1)
            },
            "flash_footprint": {
                "format": "u2",
                "shape": ("flashes",),
                "long_name": "Flash footprint size",
                "standard_name": "flash_footprint",
                "units": "L1 grid pixels",
                "default_data": lambda: np.maximum(1, np.round(RANDOM_GEN.normal(5, 3, nobs)))
            },
            "flash_id": {
                "format": "u4",
                "shape": ("flashes",),
                "long_name": "Flash footprint size",
                "standard_name": "flash_id",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "flash_time": {
                "format": "f8",
                "shape": ("flashes",),
                "long_name": "Nominal flash time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "standard_name": "time",
                "precision": "1 millisecond",
                "default_data": lambda: RANDOM_GEN.uniform(stime, etime, nobs)
            },
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ("scalar",),
                "long_name": "L1b geolocation warning",
                "default_data": lambda: -127
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ("scalar",),
                "long_name": "L1b radiometric warning",
                "default_data": lambda: -127
            },
            "number_of_events": {
                "format": "u2",
                "shape": ("flashes",),
                "long_name": "Number of events in each flash",
                "default_data": lambda: rand_type(nobs, np.uint16)
            },
            "number_of_groups": {
                "format": "u2",
                "shape": ("flashes",),
                "long_name": "Number of flashes in each flash",
                "default_data": lambda: rand_type(nobs, np.uint16)
            },
        }
    }


def l2_af_schema(settings=None):
    """Define schema for LI L2 AF product."""
    settings = settings or {}
    nacc = settings.get("num_accumulations", 1)
    npix = settings.get("num_pixels", 1234)
    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", ""),
        "dimensions": accumulation_dimensions(nacc, npix),
        "variables": {
            "accumulation_offsets": {
                "format": "u4",
                "shape": ("accumulations",),
                "default_data": lambda: rand_type(nacc, np.uint32)
            },
            "accumulation_start_times": {
                "format": "f8",
                "shape": ("accumulations",),
                "long_name": "Accumulation start time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "precision": "0.001",
                "default_data": lambda: np.linspace(0.0, 1000.0, nacc)
            },
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ("accumulations",),
                "long_name": "L1b geolocation warning",
                "default_data": lambda: rand_type(nacc, np.int8)
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ("accumulations",),
                "long_name": "L1b radiometric warning",
                "default_data": lambda: rand_type(nacc, np.int8)
            },
            "average_flash_qa": {
                "format": "u1",
                "shape": ("accumulations",),
                "default_data": lambda: rand_type(nacc, np.uint8),
                "fill_value": 255,
                "scale_factor": 0.004,
                "add_offset": 0.0,
                "long_name": "average flash confidence value",

            },
            "flash_accumulation": {
                "format": "u2",
                "shape": ("pixels",),
                "fill_value": 65535,
                "scale_factor": 0.001,
                "long_name": "Per area accumulation of flashes",
                "grid_mapping": "mtg_geos_projection",
                "units": "flashes/pixel",
                "coordinate": "sparse: x y" ,
                "default_data": lambda: np.clip(np.round(RANDOM_GEN.normal(1, 2, npix)), 1, 2 ** 16 - 1)
            },
            "mtg_geos_projection": mtg_geos_projection(),
            "x": fci_grid_definition("X", npix),
            "y": fci_grid_definition("Y", npix),
        }
    }

def l2_afr_schema(settings=None):
    """Define schema for LI L2 AFR product."""
    settings = settings or {}
    nacc = settings.get("num_accumulations", 1)
    npix = settings.get("num_pixels", 1234)

    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", ""),
        "dimensions": accumulation_dimensions(nacc, npix),
        "variables": {
            "accumulation_offsets": {
                "format": "u4",
                "shape": ("accumulations",),
                "default_data": lambda: rand_type(nacc, np.uint32)
            },
            "accumulation_start_times": {
                "format": "f8",
                "shape": ("accumulations",),
                "long_name": "Accumulation start time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "precision" : "0.001",
                "default_data": lambda: np.linspace(0.0, 1000.0, nacc)
            },
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ("accumulations",),
                "long_name": "L1b geolocation warning",
                "default_data": lambda: rand_type(nacc, np.int8)
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ("accumulations",),
                "long_name": "L1b radiometric warning",
                "default_data": lambda: rand_type(nacc, np.int8)
            },
            "average_flash_qa": {
                "format": "u1",
                "shape": ("accumulations",),
                "default_data": lambda: rand_type(nacc, np.uint8),
                "fill_value": 255,
                "scale_factor": 0.004,
                "add_offset": 0.0,
                "long_name":"average flash confidence value",

            },
            "flash_radiance": {
                "format": "u2",
                "shape": ("pixels",),
                "fill_value": 65535,
                "scale_factor": 1.0,
                "add_offset": 0.0,
                "long_name": "Area averaged flash radiance accumulation",
                "grid_mapping": "mtg_geos_projection",
                "units": "mW.m-2.sr-1",
                "coordinate": "sparse: x y" ,
                "default_data": lambda: RANDOM_GEN.integers(low=1, high=6548, size=(npix), dtype=np.int16)
            },
            "mtg_geos_projection": mtg_geos_projection(),
            "x": fci_grid_definition("X", npix),
            "y": fci_grid_definition("Y", npix),
        }
    }

def l2_afa_schema(settings=None):
    """Define schema for LI L2 AFA product."""
    settings = settings or {}
    nacc = settings.get("num_accumulations", 1)
    npix = settings.get("num_pixels", 1234)

    return {
        "providers": settings.get("providers", {}),
        "variable_path": settings.get("variable_path", ""),
        "dimensions": accumulation_dimensions(nacc, npix),
        "variables": {
            "accumulation_offsets": {
                "format": "u4",
                "shape": ("accumulations",),
                "default_data": lambda: rand_type(nacc, np.uint32)
            },
            "accumulation_start_times": {
                "format": "f8",
                "shape": ("accumulations",),
                "long_name": "Accumulation start time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "precision" : "0.001",
                "default_data": lambda: np.linspace(0.0, 1000.0, nacc)
            },
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ("accumulations",),
                "long_name": "L1b geolocation warning",
                "default_data": lambda: rand_type(nacc, np.int8)
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ("accumulations",),
                "long_name": "L1b radiometric warning",
                "default_data": lambda: rand_type(nacc, np.int8)
            },
            "average_flash_qa": {
                "format": "u1",
                "shape": ("accumulations",),
                "default_data": lambda: rand_type(nacc, np.uint8),
                "fill_value": 255,
                "scale_factor": 0.004,
                "add_offset": 0.0,
                "long_name":"average flash confidence value",

            },
            "accumulated_flash_area": {
                "format": "u4",
                "shape": ("pixels",),
                "long_name": "Number of contributing unique flashes to each pixel",
                "grid_mapping": "mtg_geos_projection",
                "coordinate": "sparse: x y" ,
                "default_data": lambda: np.mod(np.arange(npix), 10) + 1
            },
            "mtg_geos_projection": mtg_geos_projection(),
            "x": fci_grid_definition("X", npix),
            "y": fci_grid_definition("Y", npix),
        }
    }


def accumulation_dimensions(nacc, nobs):
    """Set dimensions for the accumulated products."""
    return {
        "accumulations": nacc,
        "pixels": nobs,
    }


def fci_grid_definition(axis, nobs):
    """FCI grid definition on X or Y axis."""
    scale_factor = 5.58871526031607e-5
    add_offset = -0.15561777642350116
    if axis == "X":
        long_name = "azimuth angle encoded as column"
        standard_name = "projection_x_coordinate"
        scale_factor *= -1
        add_offset *= -1
    else:
        long_name = "zenith angle encoded as row"
        standard_name = "projection_y_coordinate"

    return {
        "format": "i2",
        "shape": ("pixels",),
        "add_offset": add_offset,
        "axis": axis,
        "long_name": long_name,
        "scale_factor": scale_factor,
        "standard_name": standard_name,
        "units": "radian",
        "valid_range": np.asarray([1, 5568]),
        "default_data": lambda: np.clip(np.round(RANDOM_GEN.normal(2000, 500, nobs)), 1, 2 ** 16 - 1)
    }


def mtg_geos_projection():
    """MTG geos projection definition."""
    return {
        "format": "i4",
        "shape": ("accumulations",),
        "grid_mapping_name": "geostationary",
        "inverse_flattening": 298.257223563,
        "latitude_of_projection_origin": 0,
        "longitude_of_projection_origin": 0,
        "perspective_point_height": 3.57864e7,
        "semi_major_axis": 6378137.0,
        "semi_minor_axis": 6356752.31424518,
        "sweep_angle_axis": "y",
        "long_name": "MTG geostationary projection",
        "default_data": lambda: -2147483647
    }

#Dict containing the expecteded dtype output for each variable
expected_product_dtype = {
    "2-LE": {
        "event_id": np.uint32,
        "group_id": np.uint32,
        "l1b_chunk_ids": np.uint32,
        "l1b_chunk_offsets": np.uint32,
        "l1b_window": np.uint32,
        "filter_values": np.float32,
        "flash_id": np.uint32,
        "time_offset": np.dtype("timedelta64[ns]"),
        "epoch_time": np.dtype("datetime64[ns]"),
        "detector": np.float32,
        "l1b_filter_qa": np.float32,
        "l2_group_filter_qa": np.float32,
    },
    "2-LEF": {
        "l1b_geolocation_warning": np.int8,
        "l1b_radiometric_warning": np.int8,
        "l1b_missing_warning": np.int8,
        "event_id": np.uint32,
        "group_id": np.uint32,
        "flash_id": np.uint32,
        "detector": np.float32,
        "latitude": np.float32,
        "longitude": np.float32,
        "radiance": np.uint16,
        "event_filter_qa": np.float32,
        "epoch_time": np.dtype("datetime64[ns]"),
        "time_offset": np.dtype("timedelta64[ns]"),
        "detector_row": np.float32,
        "detector_column": np.float32,
    },
    "2-LGR": {
        "group_time": np.dtype("datetime64[ns]"),
        "l1b_geolocation_warning": np.int8,
        "l1b_radiometric_warning": np.int8,
        "latitude": np.float32,
        "longitude": np.float32,
        "radiance": np.float32,
        "group_id": np.uint32,
        "flash_id": np.uint32,
        "number_of_events": np.uint16,
        "group_filter_qa": np.float32,
    },
    "2-LFL": {
        "latitude": np.float32,
        "longitude": np.float32,
        "radiance": np.float32,
        "flash_duration": np.dtype("timedelta64[ns]"),
        "flash_filter_confidence": np.float32,
        "flash_footprint": np.uint16,
        "flash_id": np.uint32,
        "flash_time": np.dtype("datetime64[ns]"),
        "l1b_geolocation_warning": np.int8,
        "l1b_radiometric_warning": np.int8,
        "l1b_missing_warning": np.int8,
        "number_of_events": np.uint16,
        "number_of_groups": np.uint16,
    },
    "2-AF": {
        "l1b_geolocation_warning": np.int8,
        "l1b_radiometric_warning": np.int8,
        "accumulation_offsets": np.uint32,
        "accumulation_start_times": np.dtype("datetime64[ns]"),
        "average_flash_qa": np.float32,
        "mtg_geos_projection": np.int32,
        "latitude": np.float32,
        "longitude": np.float32,
        "x": np.float64,
        "y": np.float64,
        "flash_accumulation": np.float32,
    },
    "2-AFA": {
        "l1b_geolocation_warning": np.int8,
        "l1b_radiometric_warning": np.int8,
        "accumulation_offsets": np.uint32,
        "accumulation_start_times": np.dtype("datetime64[ns]"),
        "average_flash_qa": np.float32,
        "mtg_geos_projection": np.int32,
        "latitude": np.float32,
        "longitude": np.float32,
        "x": np.float64,
        "y": np.float64,
        "accumulated_flash_area": np.uint32,
    },
    "2-AFR": {
        "l1b_geolocation_warning": np.int8,
        "l1b_radiometric_warning": np.int8,
        "l1b_missing_warning": np.int8,
        "accumulation_offsets": np.uint32,
        "accumulation_start_times": np.dtype("datetime64[ns]"),
        "latitude": np.float32,
        "longitude": np.float32,
        "average_flash_qa": np.float32,
        "mtg_geos_projection": np.int32,
        "x": np.float64,
        "y": np.float64,
        "flash_radiance": np.float32,
    },
}


products_dict = {
    "2-LE": {"ftype": "li_l2_le_nc", "schema": l2_le_schema},
    "2-LEF": {"ftype": "li_l2_lef_nc", "schema": l2_lef_schema},
    "2-LGR": {"ftype": "li_l2_lgr_nc", "schema": l2_lgr_schema},
    "2-LFL": {"ftype": "li_l2_lfl_nc", "schema": l2_lfl_schema},
    "2-AF": {"ftype": "li_l2_af_nc", "schema": l2_af_schema},
    "2-AFA": {"ftype": "li_l2_afa_nc", "schema": l2_afa_schema},
    "2-AFR": {"ftype": "li_l2_afr_nc", "schema": l2_afr_schema},
}


def get_product_schema(pname, settings=None):
    """Retrieve an LI product schema given its name."""
    return products_dict[pname]["schema"](settings)


def extract_filetype_info(filetype_infos, filetype):
    """Extract Satpy-conform filetype_info from filetype_infos fixture."""
    ftype_info = filetype_infos[filetype]
    ftype_info["file_type"] = filetype
    return ftype_info


def set_variable_path(var_path, desc, sname):
    """Replace variable default path if applicable and ensure trailing separator."""
    vpath = desc.get("path", var_path)
    # Ensure we have a trailing separator:
    if vpath != "" and vpath[-1] != "/":
        vpath += "/"
    if sname != "":
        vpath += sname + "/"
    return vpath


def populate_dummy_data(data, names, details):
    """Populate variable with dummy data."""
    vname, sname = names
    desc, providers, settings = details
    if vname in providers:
        prov = providers[vname]
        # prov might be a function or directly an array that we assume will be of the correct shape:
        data[:] = prov(vname, sname, settings) if callable(prov) else prov
    else:
        # Otherwise we write the default data:
        if data.shape == ():
            # scalar case
            data = desc["default_data"]()
        else:
            data[:] = desc["default_data"]()


def add_attributes(attribs, ignored_attrs, desc):
    """Add all the custom properties directly as attributes."""
    for key, val in desc.items():
        if key not in ignored_attrs:
            attribs[key] = val


# Note: the helper class below has some missing abstract class implementation,
# but that is not critical to us, so ignoring them for now.
class FakeLIFileHandlerBase(FakeNetCDF4FileHandler):  # pylint: disable=abstract-method
    """Class for faking the NetCDF4 Filehandler."""

    # Optional parameter that may be provided at the time of the creation of this file handler
    # to customize the generated content. This may be either a simple dictionary or a callable
    # if a callable is provided it will be called to retrieve the actual parameter to be used:
    schema_parameters = None

    def get_variable_writer(self, dset, settings):
        """Get a variable writer."""
        # use a variable path prefix:
        var_path = settings.get("variable_path", "")

        # Also keep track of the potential providers:
        providers = settings.get("providers", {})

        # list of ignored attribute names:
        ignored_attrs = ["path", "format", "shape", "default_data", "fill_value"]

        # dictionary of dimensions:
        dims = settings.get("dimensions", {})

        def write_variable(vname, desc, sname=""):
            """Write a variable in our dataset."""
            # get numeric shape:
            shape_str = desc["shape"]
            shape = tuple([dims[dname] for dname in shape_str])

            # Get the desired data type:
            dtype = TYPE_MAP[desc["format"]]

            # Prepare a numpy array with the appropriate shape and type:
            data = np.zeros(shape, dtype=dtype)

            # Replace variable default path if applicable:
            vpath = set_variable_path(var_path, desc, sname)

            # Variable full name:
            full_name = f"{vpath}{vname}"

            # Add all the custom properties directly as attributes:
            attribs = {}
            add_attributes(attribs, ignored_attrs, desc)

            # Rename the fill value attribute:
            if "fill_value" in desc:
                attribs["_FillValue"] = desc["fill_value"]

            names = [vname, sname]
            details = [desc, providers, settings]
            populate_dummy_data(data, names, details)

            # Now we assign that data array:
            dset[full_name] = xr.DataArray(data, dims=shape_str, attrs=attribs)

            # Write the copy of the content:
            self.content[full_name] = data

        return write_variable

    def get_test_content(self, filename, filename_info, filetype_info):
        """Get the content of the test data.

        Here we generate the default content we want to provide depending
        on the provided filename infos.
        """
        # Retrieve the correct schema to write with potential customization parameters:
        params = FakeLIFileHandlerBase.schema_parameters
        if callable(params):
            # Note: params *IS* callable below:
            params = params(filename, filename_info, filetype_info)  # pylint: disable=not-callable

        settings = get_product_schema(filetype_info["file_desc"]["product_type"], params)

        # Resulting dataset:
        dset = {}

        # Also keep a copy of the written content:
        self.content = {}

        # Retrieve the variable writer function
        write_variable = self.get_variable_writer(dset, settings)

        # Write all the raw (i.e not in sectors) variables:
        self.write_variables(settings, write_variable)

        # Write the sector variables:
        self.write_sector_variables(settings, write_variable)

        return dset

    def write_variables(self, settings, write_variable):
        """Write raw (i.e. not in sectors) variables."""
        if "variables" in settings:
            variables = settings.get("variables")
            for vname, desc in variables.items():
                write_variable(vname, desc)

    def write_sector_variables(self, settings, write_variable):
        """Write the sector variables."""
        if "sector_variables" in settings:
            sector_vars = settings.get("sector_variables")
            sectors = settings.get("sectors", ["north", "east", "south", "west"])

            for sname in sectors:
                for vname, desc in sector_vars.items():
                    write_variable(vname, desc, sname)



"""Tests for the EarthCARE MSI L1c reader."""

import dask.array as da
import h5py
import numpy as np
import pytest
import xarray as xr

from satpy import Scene
from satpy.tests.utils import make_dataid

N_BANDS = 7
N_SCANS = 20
N_COLS = 384
SHAPE_SC = (300, 6000)
SOL_IRRAD = np.array([30.9, 19.59, 14.77, 8.25], dtype=np.float32)
# Create 384 values for SOL_IRRAD
SOL_IRRAD = SOL_IRRAD.reshape((-1, 1)) * np.ones(N_COLS, dtype=np.float32).reshape((1, -1))
DIMLIST = np.ones((N_BANDS, N_SCANS, N_COLS))

def _setup_science_data(N_BANDS, N_SCANS, N_COLS):
    # Set some default attributes
    data = {
        "pixel_values":
        xr.DataArray(
            da.ones((N_BANDS, N_SCANS, N_COLS), chunks=1024, dtype=np.float32),
            attrs={"units": "Wm-2 sr-1 or K", "DIMENSION_LIST": DIMLIST},
            dims=("band", "dim_2", "dim_1")),
        "land_flag":
        xr.DataArray(
            da.ones((N_SCANS, N_COLS), chunks=1024, dtype=np.uint16),
            attrs={"units": ""},
            dims=("along_track", "across_track")),
        "solar_azimuth_angle":
        xr.DataArray(
            da.ones((N_SCANS, N_COLS), chunks=1024, dtype=np.float32),
            attrs={"units": "degrees"},
            dims=("along_track", "across_track")),
        "longitude":
        xr.DataArray(
            da.ones((N_SCANS, N_COLS), chunks=1024, dtype=np.float32),
            attrs={"units": "degrees"},
            dims=("along_track", "across_track")),
        "latitude":
        xr.DataArray(
            da.ones((N_SCANS, N_COLS), chunks=1024, dtype=np.float32),
            attrs={"units": "degrees"},
            dims=("along_track", "across_track")),
        "solar_spectral_irradiance":
        xr.DataArray(
            da.array(SOL_IRRAD),
            attrs={"units": "W m-2"},
            dims=("band", "across_track")),
    }

    return data


@pytest.fixture(scope="session")
def msi_ec_l1c_dummy_file(tmp_path_factory):
    """Create a fake insat MSI 1C file."""
    filename = tmp_path_factory.mktemp("data") / "ECA_EXAA_MSI_RGR_1C_20250625T005649Z_20250625T024013Z_42043E.h5"

    with h5py.File(filename, "w") as fid:
        ScienceData_group = fid.create_group("ScienceData")
        for name, value in _setup_science_data(N_BANDS, N_SCANS, N_COLS).items():
            ScienceData_group[name] = value

    return filename


def test_get_pixvalues(msi_ec_l1c_dummy_file):
        """Test loadingpixel values from file."""
        res = Scene(reader="msi_l1c_earthcare", filenames=[msi_ec_l1c_dummy_file])

        available_datasets = list(res.available_dataset_ids())
        assert len(available_datasets) == 27

        res.load(["VIS", "VNIR", "TIR1", "TIR3", "solar_azimuth_angle", "land_water_mask"])
        #assert len(res) == 6
        with pytest.raises(KeyError):
            res["TIR2"]
        with pytest.raises(KeyError):
            res["SWIR1"]

        assert res["VIS"].shape == (20, N_COLS)
        assert res["VIS"].attrs["calibration"] == "reflectance"
        assert res["VIS"].attrs["units"] == "%"

        assert res["TIR1"].shape == (20, N_COLS)
        assert res["TIR1"].attrs["calibration"] == "brightness_temperature"
        assert res["TIR1"].attrs["units"] == "K"
        assert res["TIR1"].dtype == np.float32

        assert res["solar_azimuth_angle"].shape == (20, N_COLS)
        assert res["solar_azimuth_angle"].attrs["units"] == "degrees"
        assert res["solar_azimuth_angle"].dtype == np.float32

        assert res["land_water_mask"].shape == (20, N_COLS)
        assert res["land_water_mask"].attrs["units"] == 1
        assert res["land_water_mask"].dtype == np.uint16


def test_calibration(msi_ec_l1c_dummy_file):
    """Test loadingpixel values from file."""
    res = Scene(reader="msi_l1c_earthcare", filenames=[msi_ec_l1c_dummy_file])

    with pytest.raises(KeyError):
        res.load([make_dataid(name="VIS", calibration="counts")])
    with pytest.raises(KeyError):
        res.load([make_dataid(name="TIR1", calibration="counts")])
    with pytest.raises(KeyError):
        res.load([make_dataid(name="TIR1", calibration="radiance")])

    res.load([make_dataid(name="VIS", calibration="radiance")])
    assert res["VIS"].attrs["calibration"] == "radiance"
    assert res["VIS"].attrs["units"] == "W m-2 sr-1"
    assert np.all(np.array(res["VIS"].data) == 1)

    res.load([make_dataid(name="VNIR", calibration="reflectance")])
    assert res["VNIR"].attrs["calibration"] == "reflectance"
    assert res["VNIR"].attrs["units"] == "%"
    assert np.all(np.array(res["VNIR"].data) == 1 * np.pi * 100 / SOL_IRRAD[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for writing GeoTIFF files with NinJoTIFF tags."""

import datetime
import logging
import os
from unittest.mock import Mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import create_area_def
from trollimage.xrimage import XRImage

from satpy import Scene
from satpy.enhancements.enhancer import get_enhanced_image

try:
    from math import prod
except ImportError:  # Remove when dropping Python < 3.8
    from functools import reduce
    from operator import mul

    def prod(iterable):  # type: ignore
        """Drop-in replacement for math.prod."""
        return reduce(mul, iterable, 1)

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - monkeypatch
# - caplog


def _get_fake_da(lo, hi, shp, dtype="f4"):
    """Generate dask array with synthetic data.

    This is more or less a 2d linspace: it'll return a 2-d dask array of shape
    ``shp``, lowest value is ``lo``, highest value is ``hi``.
    """
    return da.linspace(lo, hi, prod(shp), dtype=dtype).reshape(shp)


@pytest.fixture(scope="module")
def test_area_tiny_eqc_sphere():
    """Create 10x00 test equirectangular area centered on (40, -30), spherical geoid, m."""
    shp = (10, 20)
    test_area = create_area_def(
        "test-area-eqc-sphere",
        {"proj": "eqc", "lat_ts": 0., "lat_0": 0., "lon_0": 0.,
         "x_0": 0., "y_0": 0., "ellps": "sphere", "units": "m",
         "no_defs": None, "type": "crs"},
        units="m",
        shape=shp,
        resolution=1000,
        center=(-3330000.0, 4440000.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_small_eqc_wgs84():
    """Create 50x100 test equirectangular area centered on (50, 90), wgs84."""
    shp = (50, 100)
    test_area = create_area_def(
            "test-area-eqc-wgs84",
            {"proj": "eqc", "lat_0": 2.5, "lon_0": 1., "ellps": "WGS84"},
            units="m",
            shape=shp,
            resolution=1000,
            center=(10000000.0, 6000000.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_tiny_stereographic_wgs84():
    """Create a 20x10 test stereographic area centered near the north pole, wgs84."""
    shp = (20, 10)
    test_area = create_area_def(
        "test-area-north-stereo",
        {"proj": "stere", "lat_0": 75.0, "lon_0": 2.0, "lat_ts": 60.0,
            "ellps": "WGS84", "units": "m", "type": "crs"},
        units="m",
        shape=shp,
        resolution=1000,
        center=(0.0, 1500000.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_tiny_antarctic():
    """Create a 20x10 test stereographic area centered near the south pole, wgs84."""
    shp = (20, 10)
    test_area = create_area_def(
        "test-area-south-stereo",
        {"proj": "stere", "lat_0": -75.0, "lon_0": 2.0, "lat_ts": 60.0,
            "ellps": "WGS84", "units": "m", "type": "crs"},
        units="m",
        shape=shp,
        resolution=1000,
        center=(0.0, -1500000.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_northpole():
    """Create a 20x10 test area centered exactly on the north pole.

    This has no well-defined central meridian so needs separate testing.
    """
    shp = (20, 10)
    test_area = create_area_def(
            "test-area-north-pole",
            {"proj": "stere", "lat_0": 90, "lat_ts": 60, "ellps": "WGS84"},
            shape=shp,
            resolution=1000,
            center=(0.0, 15000000.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_merc():
    """Create a mercator area."""
    from pyproj import CRS
    shp = (20, 10)
    test_area = create_area_def(
        "test-area-merc",
        CRS("+proj=merc"),
        units="m",
        shape=shp,
        resolution=1000,
        center=(0.0, 0.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_weird():
    """Create a weird area (interrupted goode homolosine) to test error handling."""
    from pyproj import CRS
    shp = (20, 10)
    test_area = create_area_def(
        "test-area-north-stereo",
        CRS("+proj=igh"),
        units="m",
        shape=shp,
        resolution=1000,
        center=(0.0, 1500000.0))
    return test_area


@pytest.fixture(scope="module")
def test_area_epsg4326():
    """Test with EPSG4326 (latlong) area, which has no CRS coordinate operation."""
    from pyproj import CRS
    shp = (16, 8)
    euro4326 = create_area_def(
        "epgs4326europa",
        CRS.from_epsg(4326),
        resolution=1/128,
        shape=shp,
        center=(0, 0))
    return euro4326


@pytest.fixture(scope="module")
def test_image_small_mid_atlantic_L(test_area_tiny_eqc_sphere):
    """Get a small test image in mode L, over Atlantic."""
    arr = xr.DataArray(
        _get_fake_da(-80, 40, test_area_tiny_eqc_sphere.shape + (1,)),
        dims=("y", "x", "bands"),
        attrs={
            "name": "test-small-mid-atlantic",
            "start_time": datetime.datetime(1985, 8, 13, 13, 0),
            "area": test_area_tiny_eqc_sphere},
        coords={"bands": ["L"]},
    )
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_small_mid_atlantic_K_L(test_area_tiny_eqc_sphere):
    """Get a small test image in units K, mode L, over Atlantic."""
    arr = xr.DataArray(
        _get_fake_da(-80+273.15, 40+273.15, test_area_tiny_eqc_sphere.shape + (1,)),
        dims=("y", "x", "bands"),
        attrs={
            "name": "test-small-mid-atlantic",
            "start_time": datetime.datetime(1985, 8, 13, 13, 0),
            "area": test_area_tiny_eqc_sphere,
            "units": "K"},
        coords={"bands": ["L"]},
    )
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_small_mid_atlantic_L_no_quantity(test_area_tiny_eqc_sphere):
    """Get a small test image, mode L, over Atlantic, with non-quantitywvalues.

    This could be the case, for example, for vis_with_night_ir.
    """
    arr = xr.DataArray(
        _get_fake_da(0, 273, test_area_tiny_eqc_sphere.shape + (1,)),
        dims=("y", "x", "bands"),
        coords={"bands": ["L"]},
        attrs={
            "name": "test-small-mid-atlantic",
            "start_time": datetime.datetime(1985, 8, 13, 13, 0),
            "area": test_area_tiny_eqc_sphere,
            "units": "N/A"})
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_large_asia_RGB(test_area_small_eqc_wgs84):
    """Get a large-ish test image in mode RGB, over Asia."""
    arr = xr.DataArray(
        _get_fake_da(0, 255, test_area_small_eqc_wgs84.shape + (3,), "uint8"),
        dims=("y", "x", "bands"),
        coords={"bands": ["R", "G", "B"]},
        attrs={
            "name": "test-large-asia",
            "start_time": datetime.datetime(2015, 10, 21, 20, 25, 0),
            "area": test_area_small_eqc_wgs84,
            "mode": "RGB"})
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_small_arctic_P(test_area_tiny_stereographic_wgs84):
    """Get a small-ish test image in mode P, over Arctic."""
    arr = xr.DataArray(
        _get_fake_da(0, 10, test_area_tiny_stereographic_wgs84.shape + (1,), "uint8"),
        dims=("y", "x", "bands"),
        coords={"bands": ["P"]},
        attrs={
            "name": "test-small-arctic",
            "start_time": datetime.datetime(2027, 8, 2, 8, 20),
            "area": test_area_tiny_stereographic_wgs84,
            "mode": "P"})
    # simulate an enhancement history such as palettize may add
    arr.attrs["enhancement_history"] = [
            {"scale": np.float64(0.01),
            "offset": np.float64(0.0),
            "colormap": Mock()}]

    return XRImage(arr)


@pytest.fixture(scope="module")
def test_image_northpole(test_area_northpole):
    """Test image with area exactly on northpole."""
    arr = xr.DataArray(
        _get_fake_da(1, 100, test_area_northpole.shape + (1,), "uint8"),
        dims=("y", "x", "bands"),
        coords={"bands": ["L"]},
        attrs={
            "name": "test-northpole",
            "start_time": datetime.datetime(1926, 5, 12, 0),
            "area": test_area_northpole,
            "mode": "L"})
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_weird(test_area_weird):
    """Get a small image with some weird properties to test error handling."""
    da = xr.DataArray(
        _get_fake_da(1, 2, test_area_weird.shape + (2,), "uint8"),
        dims=("y", "x", "bands"),
        coords={"bands": ["L", "A"]},
        attrs={
            "name": "interrupted image",
            "start_time": datetime.datetime(1970, 1, 1),
            "area": test_area_weird,
            "mode": "LA"})
    return get_enhanced_image(da)


@pytest.fixture(scope="module")
def test_image_rgba_merc(test_area_merc):
    """Get a small test image in mode RGBA and mercator."""
    arr = xr.DataArray(
        _get_fake_da(-80, 40, test_area_merc.shape + (4,)),
        dims=("y", "x", "bands"),
        coords={"bands": ["R", "G", "B", "A"]},
        attrs={
            "name": "test-rgba",
            "start_time": datetime.datetime(2013, 2, 22, 12, 0),
            "area": test_area_merc,
            "mode": "RGBA"})
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_cmyk_antarctic(test_area_tiny_antarctic):
    """Get a small test image in mode CMYK on south pole."""
    arr = xr.DataArray(
        _get_fake_da(-80, 40, test_area_tiny_antarctic.shape + (4,)),
        dims=("y", "x", "bands"),
        coords={"bands": ["C", "M", "Y", "K"]},
        attrs={
            "name": "test-cmyk",
            "start_time": datetime.datetime(2065, 11, 22, 11),
            "area": test_area_tiny_antarctic,
            "mode": "CMYK"})
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def test_image_latlon(test_area_epsg4326):
    """Get image with latlon areadefinition."""
    arr = xr.DataArray(
        _get_fake_da(-50, 30, test_area_epsg4326.shape + (1,)),
        dims=("y", "x", "bands"),
        coords={"bands": ["L"]},
        attrs={
            "name": "test-latlon",
            "start_time": datetime.datetime(2001, 1, 1, 0),
            "area": test_area_epsg4326,
            "mode": "L"})
    return get_enhanced_image(arr)


@pytest.fixture(scope="module")
def ntg1(test_image_small_mid_atlantic_L):
    """Create instance of NinJoTagGenerator class."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_small_mid_atlantic_L,
            255,
            "quinoa.tif",
            ChannelID=900015,
            DataType="GORN",
            PhysicUnit="C",
            PhysicValue="Temperature",
            SatelliteNameID=6400014,
            DataSource="dowsing rod")


@pytest.fixture(scope="module")
def ntg2(test_image_large_asia_RGB):
    """Create instance of NinJoTagGenerator class."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_large_asia_RGB,
            0,
            "seitan.tif",
            ChannelID=1000015,
            DataType="GORN",
            PhysicUnit="N/A",
            PhysicValue="N/A",
            SatelliteNameID=6400014)


@pytest.fixture(scope="module")
def ntg3(test_image_small_arctic_P):
    """Create instance of NinJoTagGenerator class."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_small_arctic_P,
            255,
            "spelt.tif",
            ChannelID=800012,
            DataType="PPRN",
            PhysicUnit="N/A",
            PhysicValue="N/A",
            SatelliteNameID=6500014,
            OverFlightTime=42)


@pytest.fixture(scope="module")
def ntg_northpole(test_image_northpole):
    """Create NinJoTagGenerator with north pole image."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_northpole,
            255,
            "lentils.tif",
            ChannelID=900012,
            DataType="PORN",
            PhysicUnit="Temperature",
            PhysicValue="K",
            SatelliteNameID=7500014)


@pytest.fixture(scope="module")
def ntg_weird(test_image_weird):
    """Create NinJoTagGenerator instance with weird image."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_weird,
            12,
            "tempeh.tif",
            ChannelID=800012,
            DataType="PPRN",
            PhysicUnit="N/A",
            PhysicValue="N/A",
            SatelliteNameID=6500014)


@pytest.fixture(scope="module")
def ntg_no_fill_value(test_image_small_mid_atlantic_L):
    """Create instance of NinJoTagGenerator class."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_small_mid_atlantic_L,
            None,
            "bulgur.tif",
            ChannelID=900015,
            DataType="GORN",
            PhysicUnit="C",
            PhysicValue="Temperature",
            SatelliteNameID=6400014,
            DataSource="dowsing rod")


@pytest.fixture(scope="module")
def ntg_rgba(test_image_rgba_merc):
    """Create NinJoTagGenerator instance with RGBA image."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_rgba_merc,
            12,
            "soy.tif",
            ChannelID=800042,
            DataType="GORN",
            PhysicUnit="N/A",
            PhysicValue="N/A",
            SatelliteNameID=6500014)


@pytest.fixture(scope="module")
def ntg_cmyk(test_image_cmyk_antarctic):
    """Create NinJoTagGenerator instance with CMYK image."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_cmyk_antarctic,
            0,
            "tvp.tif",
            ChannelID=123042,
            DataType="PPRN",
            PhysicUnit="N/A",
            PhysicValue="N/A",
            SatelliteNameID=6500014)


@pytest.fixture(scope="module")
def ntg_latlon(test_image_latlon):
    """Create NinJoTagGenerator with latlon-area image."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    return NinJoTagGenerator(
            test_image_latlon,
            0,
            "latlon.tif",
            ChannelID=123456,
            DataType="GORN",
            PhysicUnit="%",
            PhysicValue="Reflectance",
            SatelliteNameID=654321)


@pytest.fixture
def _patch_datetime_now(monkeypatch):
    """Get a fake datetime.datetime.now()."""
    # Source: https://stackoverflow.com/a/20503374/974555, CC-BY-SA 4.0

    class mydatetime(datetime.datetime):
        """Drop-in replacement for datetime.datetime."""

        @classmethod
        def now(cls, tz=datetime.timezone.utc):
            """Drop-in replacement for datetime.datetime.now."""
            return datetime.datetime(2033, 5, 18, 3, 33, 20,
                                     tzinfo=tz)

    monkeypatch.setattr(datetime, "datetime", mydatetime)


def test_write_and_read_file(test_image_small_mid_atlantic_L, tmp_path):
    """Test that it writes a GeoTIFF with the appropriate NinJo-tags."""
    import rasterio

    from satpy.writers.ninjogeotiff import NinJoGeoTIFFWriter
    fn = os.fspath(tmp_path / "test.tif")
    ngtw = NinJoGeoTIFFWriter()
    ngtw.save_dataset(
        test_image_small_mid_atlantic_L.data,
        filename=fn,
        fill_value=0,
        blockxsize=128,
        blockysize=128,
        compress="lzw",
        predictor=2,
        PhysicUnit="C",
        PhysicValue="Temperature",
        SatelliteNameID=6400014,
        ChannelID=900015,
        DataType="GORN",
        DataSource="dowsing rod")
    src = rasterio.open(fn)
    tgs = src.tags()
    assert tgs["ninjo_FileName"] == fn
    assert tgs["ninjo_DataSource"] == "dowsing rod"
    np.testing.assert_allclose(float(tgs["ninjo_Gradient"]),
                               0.46771654391851947)
    np.testing.assert_allclose(float(tgs["ninjo_AxisIntercept"]),
                               -79.86771951938239)


def test_write_and_read_file_RGB(test_image_large_asia_RGB, tmp_path):
    """Test writing and reading RGB."""
    import rasterio

    from satpy.writers.ninjogeotiff import NinJoGeoTIFFWriter
    fn = os.fspath(tmp_path / "test.tif")
    ngtw = NinJoGeoTIFFWriter()
    ngtw.save_dataset(
        test_image_large_asia_RGB.data,
        filename=fn,
        fill_value=0,
        PhysicUnit="N/A",
        PhysicValue="N/A",
        SatelliteNameID=6400014,
        ChannelID=900015,
        DataType="GORN",
        DataSource="dowsing rod")
    src = rasterio.open(fn)
    tgs = src.tags()
    assert tgs["ninjo_FileName"] == fn
    assert tgs["ninjo_DataSource"] == "dowsing rod"
    assert "ninjo_Gradient" not in tgs.keys()
    assert "ninjo_AxisIntercept" not in tgs.keys()
    assert tgs["ninjo_PhysicValue"] == "N/A"


def test_write_and_read_file_LA(test_image_latlon, tmp_path):
    """Test writing and reading LA image."""
    import rasterio

    from satpy.writers.ninjogeotiff import NinJoGeoTIFFWriter
    fn = os.fspath(tmp_path / "test.tif")
    ngtw = NinJoGeoTIFFWriter()
    ngtw.save_dataset(
        test_image_latlon.data,
        filename=fn,
        fill_value=None,  # to make it LA
        PhysicUnit="%",
        PhysicValue="Reflectance",
        SatelliteNameID=6400014,
        ChannelID=900015,
        DataType="GORN",
        DataSource="dowsing rod")
    src = rasterio.open(fn)
    assert len(src.indexes) == 2  # mode LA
    tgs = src.tags()
    assert tgs["ninjo_FileName"] == fn
    assert tgs["ninjo_DataSource"] == "dowsing rod"
    np.testing.assert_allclose(float(tgs["ninjo_Gradient"]), 0.31058823679007746)
    np.testing.assert_allclose(float(tgs["ninjo_AxisIntercept"]), -49.6)
    assert tgs["ninjo_PhysicValue"] == "Reflectance"
    assert tgs["ninjo_TransparentPixel"] == "-1"  # meaning not set


def test_write_and_read_file_P(test_image_small_arctic_P, tmp_path):
    """Test writing and reading P image."""
    import rasterio
    from trollimage.colormap import Colormap

    from satpy.writers.ninjogeotiff import NinJoGeoTIFFWriter
    fn = os.fspath(tmp_path / "test.tif")
    ngtw = NinJoGeoTIFFWriter()
    ngtw.save_image(
        test_image_small_arctic_P,
        filename=fn,
        fill_value=255,
        PhysicUnit="satdata",
        PhysicValue="satdata",
        SatelliteNameID=6400014,
        ChannelID=900015,
        DataType="PPRN",
        DataSource="dowsing rod",
        keep_palette=True,
        cmap=Colormap(*enumerate(zip(*([np.linspace(0, 1, 256)]*3)))))
    src = rasterio.open(fn)
    assert len(src.indexes) == 1  # mode P
    assert src.colorinterp[0] == rasterio.enums.ColorInterp.palette
    tgs = src.tags()
    assert tgs["ninjo_FileName"] == fn
    assert tgs["ninjo_DataSource"] == "dowsing rod"
    assert tgs["ninjo_Gradient"] == "1.0"
    assert tgs["ninjo_AxisIntercept"] == "0.0"


def test_write_and_read_file_units(
        test_image_small_mid_atlantic_K_L, tmp_path, caplog):
    """Test that it writes a GeoTIFF with the appropriate NinJo-tags and units."""
    import rasterio

    from satpy.writers.ninjogeotiff import NinJoGeoTIFFWriter
    fn = os.fspath(tmp_path / "test.tif")
    ngtw = NinJoGeoTIFFWriter()
    with caplog.at_level(logging.DEBUG):
        ngtw.save_dataset(
            test_image_small_mid_atlantic_K_L.data,
            filename=fn,
            fill_value=0,
            blockxsize=128,
            blockysize=128,
            compress="lzw",
            predictor=2,
            PhysicUnit="C",
            PhysicValue="Temperature",
            SatelliteNameID=6400014,
            ChannelID=900015,
            DataType="GORN",
            DataSource="dowsing rod")
    assert "Adding offset for K → °C conversion" in caplog.text
    # a better test would be to check that the attributes haven't changed at
    # all, but that currently fails due to
    # https://github.com/pytroll/satpy/issues/2022
    assert test_image_small_mid_atlantic_K_L.data.attrs["enhancement_history"][0] != {"scale": 1, "offset": 273.15}
    src = rasterio.open(fn)
    tgs = src.tags()
    assert tgs["ninjo_FileName"] == fn
    assert tgs["ninjo_DataSource"] == "dowsing rod"
    np.testing.assert_allclose(float(tgs["ninjo_Gradient"]),
                               0.467717, rtol=1e-5)
    np.testing.assert_allclose(float(tgs["ninjo_AxisIntercept"]),
                               -79.86771, rtol=1e-5)
    fn2 = os.fspath(tmp_path / "test2.tif")
    with caplog.at_level(logging.WARNING):
        ngtw.save_dataset(
            test_image_small_mid_atlantic_K_L.data,
            filename=fn2,
            fill_value=0,
            blockxsize=128,
            blockysize=128,
            compress="lzw",
            predictor=2,
            PhysicUnit="F",
            PhysicValue="Temperature",
            SatelliteNameID=6400014,
            ChannelID=900015,
            DataType="GORN",
            DataSource="dowsing rod")
    assert ("Writing F to ninjogeotiff headers, but "
            "data attributes have unit K. "
            "No conversion applied.") in caplog.text


@pytest.mark.parametrize("unit", ["N/A", "1", ""])
def test_write_and_read_no_quantity(
        test_image_small_mid_atlantic_L_no_quantity, tmp_path, unit):
    """Test that no scale/offset written if no valid units present."""
    import rasterio

    from satpy.writers.ninjogeotiff import NinJoGeoTIFFWriter
    fn = os.fspath(tmp_path / "test.tif")
    ngtw = NinJoGeoTIFFWriter()
    ngtw.save_dataset(
        test_image_small_mid_atlantic_L_no_quantity.data,
        filename=fn,
        blockxsize=128,
        blockysize=128,
        compress="lzw",
        predictor=2,
        PhysicUnit=unit,
        PhysicValue="N/A",
        SatelliteNameID=6400014,
        ChannelID=900015,
        DataType="GORN",
        DataSource="dowsing rod")
    src = rasterio.open(fn)
    tgs = src.tags()
    assert "ninjo_Gradient" not in tgs.keys()
    assert "ninjo_AxisIntercept" not in tgs.keys()


def test_write_and_read_via_scene(test_image_small_mid_atlantic_L, tmp_path):
    """Test that all attributes are written also when writing from scene.

    It appears that :func:`Satpy.Scene.save_dataset` does not pass the filename
    to the writer.  Test that filename is still written to header when saving
    this way (the regular way).
    """
    import rasterio
    sc = Scene()
    fn = os.fspath(tmp_path / "test-{name}.tif")
    sc["montanha-do-pico"] = test_image_small_mid_atlantic_L.data
    sc.save_dataset(
        "montanha-do-pico",
        writer="ninjogeotiff",
        filename=fn,
        fill_value=0,
        PhysicUnit="C",
        PhysicValue="Temperature",
        SatelliteNameID=6400014,
        ChannelID=900015,
        DataType="GORN")
    src = rasterio.open(tmp_path / "test-montanha-do-pico.tif")
    tgs = src.tags()
    assert tgs["ninjo_FileName"] == os.fspath(tmp_path / "test-montanha-do-pico.tif")


def test_get_all_tags(ntg1, ntg3, ntg_latlon, ntg_northpole, caplog):
    """Test getting all tags from dataset."""
    # test that passed, dynamic, and mandatory tags are all included, and
    # nothing more
    t1 = ntg1.get_all_tags()
    assert set(t1.keys()) == (
            ntg1.fixed_tags.keys() |
            ntg1.passed_tags |
            ntg1.dynamic_tags.keys() |
            {"DataSource"})
    # test that when extra tag is passed this is also included
    t3 = ntg3.get_all_tags()
    assert t3.keys() == (
            ntg3.fixed_tags.keys() |
            ntg3.passed_tags |
            ntg3.dynamic_tags.keys() |
            {"OverFlightTime"})
    assert t3["OverFlightTime"] == 42
    # test that CentralMeridian skipped and warning logged
    with caplog.at_level(logging.DEBUG):
        t_latlon = ntg_latlon.get_all_tags()
    assert ("Unable to obtain value for optional NinJo tag CentralMeridian"
            in caplog.text)
    assert "CentralMeridian" not in t_latlon.keys()
    t_northpole = ntg_northpole.get_all_tags()
    assert "CentralMeridian" not in t_northpole.keys()


def test_calc_single_tag_by_name(ntg1, ntg2, ntg3):
    """Test calculating single tag from dataset."""
    assert ntg1.get_tag("Magic") == "NINJO"
    assert ntg1.get_tag("DataType") == "GORN"
    assert ntg2.get_tag("DataType") == "GORN"
    assert ntg3.get_tag("DataType") == "PPRN"
    assert ntg1.get_tag("DataSource") == "dowsing rod"
    with pytest.raises(ValueError, match="Unknown tag: invalid"):
        ntg1.get_tag("invalid")
    with pytest.raises(ValueError,
                       match="Optional tag OriginalHeader must be supplied by user if user wants to request the value,"
                             " but wasn't."):
        ntg1.get_tag("OriginalHeader")
    with pytest.raises(ValueError, match="Tag Gradient is added later by the GeoTIFF writer."):
        ntg1.get_tag("Gradient")


def test_get_central_meridian(ntg1, ntg2, ntg3, ntg_latlon, ntg_northpole):
    """Test calculating the central meridian."""
    cm = ntg1.get_central_meridian()
    assert isinstance(cm, float)
    np.testing.assert_allclose(cm, 0.0)
    np.testing.assert_allclose(ntg2.get_central_meridian(), 1.0)
    np.testing.assert_allclose(ntg3.get_central_meridian(), 2.0)
    with pytest.raises(AttributeError):
        # latlon area has no central meridian
        ntg_latlon.get_central_meridian()
    with pytest.raises(KeyError):
        # nor does area exactly on northpole
        ntg_northpole.get_central_meridian()


def test_get_color_depth(ntg1, ntg2, ntg3, ntg_weird, ntg_rgba, ntg_cmyk):
    """Test extracting the color depth."""
    cd = ntg1.get_color_depth()
    assert isinstance(cd, int)
    assert cd == 8  # mode L
    assert ntg2.get_color_depth() == 24  # mode RGB
    assert ntg3.get_color_depth() == 8  # mode P
    assert ntg_weird.get_color_depth() == 16  # mode LA
    assert ntg_rgba.get_color_depth() == 32  # mode RGBA
    with pytest.raises(ValueError, match="Unsupported image mode: CMYK"):
        ntg_cmyk.get_color_depth()


@pytest.mark.usefixtures("_patch_datetime_now")
def test_get_creation_date_id(ntg1, ntg2, ntg3):
    """Test getting the creation date ID.

    This is the time at which the file was created.

    This test believes it is run at 2033-5-18 05:33:20Z.
    """
    cdid = ntg1.get_creation_date_id()
    assert isinstance(cdid, int)
    assert cdid == 2000000000
    assert ntg2.get_creation_date_id() == 2000000000
    assert ntg3.get_creation_date_id() == 2000000000


def test_get_date_id(ntg1, ntg2, ntg3):
    """Test getting the date ID."""
    did = ntg1.get_date_id()
    assert isinstance(did, int)
    assert did == 492786000
    assert ntg2.get_date_id() == 1445459100
    assert ntg3.get_date_id() == 1817194800


def test_get_earth_radius_large(ntg1, ntg2, ntg3):
    """Test getting the Earth semi-major axis."""
    erl = ntg1.get_earth_radius_large()
    assert isinstance(erl, float)
    np.testing.assert_allclose(erl, 6370997.0)
    np.testing.assert_allclose(ntg2.get_earth_radius_large(), 6378137.0)
    np.testing.assert_allclose(ntg3.get_earth_radius_large(), 6378137.0)


def test_get_earth_radius_small(ntg1, ntg2, ntg3):
    """Test getting the Earth semi-minor axis."""
    ers = ntg1.get_earth_radius_small()
    assert isinstance(ers, float)
    np.testing.assert_allclose(ers, 6370997.0)
    np.testing.assert_allclose(ntg2.get_earth_radius_small(), 6356752.314245179)
    np.testing.assert_allclose(ntg3.get_earth_radius_small(), 6356752.314245179)


def test_get_filename(ntg1, ntg2, ntg3):
    """Test getting the filename."""
    assert ntg1.get_filename() == "quinoa.tif"
    assert ntg2.get_filename() == "seitan.tif"
    assert ntg3.get_filename() == "spelt.tif"


def test_get_min_gray_value_L(ntg1):
    """Test getting min gray value for mode L."""
    mg = ntg1.get_min_gray_value()
    assert isinstance(mg.compute().item(), int)
    assert mg.compute() == 0


def test_get_min_gray_value_RGB(ntg2):
    """Test getting min gray value for RGB.

    Note that min/max gray value is mandatory in NinJo even for RGBs?
    """
    assert ntg2.get_min_gray_value().compute().item() == 1  # fill value 0


def test_get_min_gray_value_P(ntg3):
    """Test getting min gray value for mode P."""
    assert ntg3.get_min_gray_value().compute().item() == 0


def test_get_max_gray_value_L(ntg1):
    """Test getting max gray value for mode L."""
    mg = ntg1.get_max_gray_value().compute().item()
    assert isinstance(mg, int)
    assert mg == 254  # fill value is 255


def test_get_max_gray_value_RGB(ntg2):
    """Test max gray value for RGB."""
    assert ntg2.get_max_gray_value() == 255


def test_get_max_gray_value_P(ntg3):
    """Test getting max gray value for mode P."""
    assert ntg3.get_max_gray_value().compute().item() == 10


@pytest.mark.xfail(reason="not easy, not needed, not implemented")
def test_get_meridian_east(ntg1, ntg2, ntg3):
    """Test getting east meridian."""
    np.testing.assert_allclose(ntg1.get_meridian_east(), -29.048101549452294)
    np.testing.assert_allclose(ntg2.get_meridian_east(), 180.0)
    np.testing.assert_allclose(ntg3.get_meridian_east(), 99.81468125314737)


@pytest.mark.xfail(reason="not easy, not needed, not implemented")
def test_get_meridian_west(ntg1, ntg2, ntg3):
    """Test getting west meridian."""
    np.testing.assert_allclose(ntg1.get_meridian_west(), -30.846745608241903)
    np.testing.assert_allclose(ntg2.get_meridian_east(), -180.0)
    np.testing.assert_allclose(ntg3.get_meridian_west(), 81.84837557075694)


def test_get_projection(ntg1, ntg2, ntg3, ntg_weird, ntg_rgba, ntg_cmyk,
                        ntg_latlon):
    """Test getting projection string."""
    assert ntg1.get_projection() == "PLAT"
    assert ntg2.get_projection() == "PLAT"
    assert ntg3.get_projection() == "NPOL"
    assert ntg_cmyk.get_projection() == "SPOL"
    assert ntg_rgba.get_projection() == "MERC"
    assert ntg_latlon.get_projection() == "PLAT"
    with pytest.raises(ValueError, match="Unknown mapping from area .*"):
        ntg_weird.get_projection()


def test_get_ref_lat_1(ntg1, ntg2, ntg3, ntg_weird, ntg_latlon):
    """Test getting reference latitude 1."""
    rl1 = ntg1.get_ref_lat_1()
    assert isinstance(rl1, float)
    np.testing.assert_allclose(rl1, 0.0)
    np.testing.assert_allclose(ntg2.get_ref_lat_1(), 2.5)
    np.testing.assert_allclose(ntg3.get_ref_lat_1(), 75)
    with pytest.raises(ValueError, match="Could not find reference latitude for area test-area-north-stereo"):
        ntg_weird.get_ref_lat_1()
    with pytest.raises(AttributeError):
        ntg_latlon.get_ref_lat_1()


@pytest.mark.xfail(reason="Not implemented, what is this?")
def test_get_ref_lat_2(ntg1, ntg2, ntg3):
    """Test getting reference latitude 2."""
    rl2 = ntg1.get_ref_lat_2()
    assert isinstance(rl2, float)
    np.testing.assert_allclose(rl2, 0.0)
    np.testing.assert_allclose(ntg2.get_ref_lat_2(), 0.0)
    np.testing.assert_allclose(ntg2.get_ref_lat_3(), 0.0)


def test_get_transparent_pixel(ntg1, ntg2, ntg3, ntg_no_fill_value):
    """Test getting fill value."""
    tp = ntg1.get_transparent_pixel()
    assert isinstance(tp, int)
    assert tp == 255
    assert ntg2.get_transparent_pixel() == 0
    assert ntg3.get_transparent_pixel() == 255
    assert ntg_no_fill_value.get_transparent_pixel() == -1


def test_get_xmax(ntg1, ntg2, ntg3):
    """Test getting maximum x."""
    xmax = ntg1.get_xmaximum()
    assert isinstance(xmax, int)
    assert xmax == 20
    assert ntg2.get_xmaximum() == 100
    assert ntg3.get_xmaximum() == 10


def test_get_ymax(ntg1, ntg2, ntg3):
    """Test getting maximum y."""
    ymax = ntg1.get_ymaximum()
    assert isinstance(ymax, int)
    assert ymax == 10
    assert ntg2.get_ymaximum() == 50
    assert ntg3.get_ymaximum() == 20


def test_create_unknown_tags(test_image_small_arctic_P):
    """Test that unknown tags raise ValueError."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    with pytest.raises(ValueError, match="The following tags were not recognised: Locatie"):
        NinJoTagGenerator(
            test_image_small_arctic_P,
            42,
            "quorn.tif",
            ChannelID=800012,
            DataType="GPRN",
            PhysicUnit="N/A",
            PhysicValue="N/A",
            SatelliteNameID=6500014,
            Locatie="Hozomeen")


def test_str_ids(test_image_small_arctic_P):
    """Test that channel and satellit IDs can be str."""
    from satpy.writers.ninjogeotiff import NinJoTagGenerator
    NinJoTagGenerator(
        test_image_small_arctic_P,
        42,
        "quorn.tif",
        ChannelID="la manche",
        DataType="GPRN",
        PhysicUnit="N/A",
        PhysicValue="N/A",
        SatelliteNameID="trollsat")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Geostationary High-speed Imager reader for the Level_1 HDF format.

This instrument is aboard the Fengyun-4B satellite. No document is available to describe this
format is available, but it's broadly similar to the co-flying AGRI instrument.

"""

import logging

from pyproj import Proj

from satpy.readers.core._geos_area import get_area_definition
from satpy.readers.core.fy4 import FY4Base

logger = logging.getLogger(__name__)


class HDF_GHI_L1(FY4Base):
    """GHI l1 file handler."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init filehandler."""
        super(HDF_GHI_L1, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "GHI"

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        ds_name = dataset_id["name"]
        logger.debug("Reading in get_dataset %s.", ds_name)
        file_key = ds_info.get("file_key", ds_name)
        if self.CHANS_ID in file_key:
            file_key = f"Data/{file_key}"
        elif self.SUN_ID in file_key or self.SAT_ID in file_key:
            file_key = f"Navigation/{file_key}"
        data = self.get(file_key)
        if data.ndim >= 2:
            data = data.rename({data.dims[-2]: "y", data.dims[-1]: "x"})

        data = self.calibrate(data, ds_info, ds_name, file_key)

        self.adjust_attrs(data, ds_info)

        return data

    def adjust_attrs(self, data, ds_info):
        """Adjust the attrs of the data."""
        satname = self.PLATFORM_NAMES.get(self["/attr/Satellite Name"], self["/attr/Satellite Name"])
        data.attrs.update({"platform_name": satname,
                           "sensor": self["/attr/Sensor Identification Code"].lower(),
                           "orbital_parameters": {
                               "satellite_nominal_latitude": self["/attr/NOMSubSatLat"].item(),
                               "satellite_nominal_longitude": self["/attr/NOMSubSatLon"].item(),
                               "satellite_nominal_altitude": self["/attr/NOMSatHeight"].item()}})
        data.attrs.update(ds_info)
        # remove attributes that could be confusing later
        data.attrs.pop("FillValue", None)
        data.attrs.pop("Intercept", None)
        data.attrs.pop("Slope", None)

    def get_area_def(self, key):
        """Get the area definition."""
        # Coordination Group for Meteorological Satellites LRIT/HRIT Global Specification
        # https://www.cgms-info.org/documents/cgms-lrit-hrit-global-specification-(v2-8-of-30-oct-2013).pdf
        res = key["resolution"]

        pdict = {}

        c_lats = self.file_content["/attr/Corner-Point Latitudes"]
        c_lons = self.file_content["/attr/Corner-Point Longitudes"]

        p1 = (c_lons[0], c_lats[0])
        p2 = (c_lons[1], c_lats[1])
        p3 = (c_lons[2], c_lats[2])
        p4 = (c_lons[3], c_lats[3])

        pdict["a"] = self.file_content["/attr/Semi_major_axis"] * 1E3  # equator radius (m)
        pdict["b"] = self.file_content["/attr/Semi_minor_axis"] * 1E3  # equator radius (m)
        pdict["h"] = self.file_content["/attr/NOMSatHeight"] * 1E3  # the altitude of satellite (m)

        pdict["h"] = pdict["h"] - pdict["a"]

        pdict["ssp_lon"] = float(self.file_content["/attr/NOMSubSatLon"])
        pdict["nlines"] = float(self.file_content["/attr/RegLength"])
        pdict["ncols"] = float(self.file_content["/attr/RegWidth"])

        pdict["scandir"] = "S2N"

        pdict["a_desc"] = "FY-4 {} area".format(self.filename_info["observation_type"])
        pdict["a_name"] = f'{self.filename_info["observation_type"]}_{res}m'
        pdict["p_id"] = f"FY-4, {res}m"

        proj_dict = {"a": pdict["a"],
                     "b": pdict["b"],
                     "lon_0": pdict["ssp_lon"],
                     "h": pdict["h"],
                     "proj": "geos",
                     "units": "m",
                     "sweep": "y"}

        p = Proj(proj_dict)
        o1 = (p(p1[0], p1[1]))  # Upper left
        o2 = (p(p2[0], p2[1]))  # Upper right
        o3 = (p(p3[0], p3[1]))  # Lower left
        o4 = (p(p4[0], p4[1]))  # Lower right

        deller = res / 2.

        area = get_area_definition(pdict, (o3[0] - deller, o4[1] - deller, o2[0], o1[1]))

        return area
